#!/usr/bin/env python3
#
# Tests writing of Python and NumPy equations.
# These are used in pyfunc() and related, so numerical tests are possible and
# desirable.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import math
import unittest

import numpy

import myokit
import myokit.formats.python

from myokit import (
    Number, PrefixPlus, PrefixMinus, Plus, Minus,
    Multiply, Divide, Quotient, Remainder, Power, Sqrt,
    Exp, Log, Log10, Sin, Cos, Tan, ASin, ACos, ATan, Floor, Ceil, Abs,
    Not, And, Or, Equal, NotEqual, More, Less, MoreEqual, LessEqual,
    If, Piecewise,
)

import myokit.tests


class PythonExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """
    Test conversion of expressions to Python.
    This is used by pyfunc(), although that's usually done with the
    NumPyExpressionWriter instead.
    Numerical tests are provided.
    """
    _name = 'python'
    _target = myokit.formats.python.PythonExpressionWriter

    def py(self, expression, expected):
        """ Test by converting to Python and executing. """
        function = expression.pyfunc(use_numpy=False)
        value = function()
        self.assertEqual(value, expected)

    def test_number(self):
        self.eq(Number(1), '1.0')
        self.eq(Number(-1.3274924373284374), '-1.32749243732843736e+00')
        self.eq(Number(+1.3274924373284374), '1.32749243732843736e+00')
        self.eq(Number(-2), '-2.0')
        self.eq(Number(13, 'mV'), '13.0')
        self.py(Number(14, 'mV'), 14)

    def test_name(self):
        self.eq(self.a, 'a')
        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.a), 'COMP.A')

    def test_derivative(self):
        self.eq(myokit.Derivative(self.a), 'dot(a)')

    def test_partial_derivative(self):
        e = myokit.PartialDerivative(self.a, self.b)
        self.assertRaisesRegex(NotImplementedError, 'Partial', self.w.ex, e)

    def test_initial_value(self):
        e = myokit.InitialValue(self.a)
        self.assertRaisesRegex(NotImplementedError, 'Initial', self.w.ex, e)

    def test_prefix_plus(self):
        # Test with numbers
        p = Number(11, 'kV')
        self.eq(PrefixPlus(p), '+11.0')
        self.eq(PrefixPlus(PrefixPlus(p)), '++11.0')
        self.eq(PrefixPlus(Number('+1')), '+1.0')
        self.py(PrefixPlus(Number(3)), 3)

        # Test with operators of precedence SUM, PRODUCT, POWER
        a, b, c = self.abc
        self.eq(PrefixPlus(Plus(a, b)), '+(a + b)')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c), '+(a + b) / c')
        self.eq(PrefixPlus(Remainder(b, a)), '+(b % a)')
        self.eq(Power(PrefixPlus(a), b), '(+a)**b')
        self.eq(Power(PrefixPlus(Power(b, a)), c), '(+b**a)**c')
        self.eq(Power(a, PrefixPlus(Power(b, c))), 'a**(+b**c)')

    def test_prefix_minus(self):
        # Test with numbers
        p = Number(11, 'kV')
        self.eq(PrefixMinus(p), '-11.0')
        self.eq(PrefixMinus(PrefixMinus(p)), '--11.0')
        self.eq(PrefixMinus(Number(-1)), '--1.0')
        self.py(PrefixMinus(Number(3)), -3)

        # Test with operators of precedence SUM, PRODUCT, POWER
        a, b, c = self.abc
        self.eq(PrefixMinus(Minus(a, b)), '-(a - b)')
        self.eq(Multiply(PrefixMinus(Plus(b, a)), c), '-(b + a) * c')
        self.eq(PrefixMinus(Quotient(b, a)), '-(b // a)')
        self.eq(Power(PrefixMinus(a), b), '(-a)**b')
        self.eq(Power(PrefixMinus(Power(c, b)), a), '(-c**b)**a')
        self.eq(Power(a, PrefixMinus(Power(b, c))), 'a**(-b**c)')

    def test_plus_minus(self):
        a, b, c = self.abc
        self.eq(Plus(a, b), 'a + b')
        self.eq(Plus(Plus(a, b), c), 'a + b + c')
        self.eq(Plus(a, Plus(b, c)), 'a + (b + c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Minus(Minus(a, b), c), 'a - b - c')
        self.eq(Minus(a, Minus(b, c)), 'a - (b - c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Plus(Minus(a, b), c), 'a - b + c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')
        self.eq(Minus(Plus(a, b), c), 'a + b - c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')

        self.py(Plus(Number(4), Number(2)), 6)
        self.py(Minus(Number(5), Number(1.5)), 3.5)

    def test_multiply_divide(self):
        a, b, c = self.abc
        self.eq(Multiply(a, b), 'a * b')
        # Left-to-right, so (a * b) * c is the same as a * b * c...
        self.eq(Multiply(Multiply(a, b), c), 'a * b * c')
        # ...but order-of-operations-wise, a * (b * c) is different!
        self.eq(Multiply(a, Multiply(b, c)), 'a * (b * c)')
        # Note that a user typing a * b * c results in (a * b) * c

        self.eq(Divide(a, b), 'a / b')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')

        self.eq(Divide(Multiply(a, b), c), 'a * b / c')
        self.eq(Multiply(Divide(a, b), c), 'a / b * c')
        self.eq(Divide(a, Multiply(b, c)), 'a / (b * c)')
        self.eq(Multiply(a, Divide(b, c)), 'a * (b / c)')

        self.eq(Multiply(Minus(a, b), c), '(a - b) * c')
        self.eq(Multiply(a, Plus(b, c)), 'a * (b + c)')
        self.eq(Minus(Multiply(a, b), c), 'a * b - c')
        self.eq(Plus(a, Multiply(b, c)), 'a + b * c')
        self.eq(Divide(Plus(a, b), c), '(a + b) / c')
        self.eq(Divide(a, Minus(b, c)), 'a / (b - c)')
        self.eq(Plus(Divide(a, b), c), 'a / b + c')
        self.eq(Minus(a, Divide(b, c)), 'a - b / c')

        self.py(Multiply(Number(7), Number(9)), 63)
        self.py(Divide(Number(5), Number(2)), 2.5)
        self.py(Multiply(Number(7), Plus(Number(1), Number(9))), 70)
        self.py(Divide(Minus(Number(19), Number(5)), Number(2)), 7)

    def test_quotient_remainder(self):
        a, b, c = self.abc

        self.eq(Remainder(Multiply(a, b), c), 'a * b % c')
        self.eq(Multiply(Quotient(a, b), c), 'a // b * c')
        self.eq(Quotient(a, Multiply(b, c)), 'a // (b * c)')
        self.eq(Multiply(a, Remainder(b, c)), 'a * (b % c)')

        self.eq(Divide(a, Quotient(b, c)), 'a / (b // c)')
        self.eq(Divide(Quotient(b, c), a), 'b // c / a')
        self.eq(Divide(a, Remainder(b, c)), 'a / (b % c)')
        self.eq(Divide(Remainder(b, c), a), 'b % c / a')

        self.py(Quotient(Number(10), Number(4)), 2)     # 2*4 + 2
        self.py(Remainder(Number(10), Number(4)), 2)
        self.py(Quotient(Number(10), Number(6)), 1)     # 6 + 4
        self.py(Remainder(Number(10), Number(6)), 4)
        self.py(Quotient(Number(5), Number(3)), 1)      # 1*3 + 2
        self.py(Remainder(Number(5), Number(3)), 2)
        self.py(Quotient(Number(-5), Number(3)), -2)    # -2*3 + 1
        self.py(Remainder(Number(-5), Number(3)), 1)
        self.py(Quotient(Number(5), Number(-3)), -2)    # -2*-3 - 1
        self.py(Remainder(Number(5), Number(-3)), -1)
        self.py(Quotient(Number(-5), Number(-3)), 1)    # 1*-3 - 2
        self.py(Remainder(Number(-5), Number(-3)), -2)
        # The quotient has sign sign(a)*sign(b)

    def test_power(self):
        a, b, c = self.abc
        self.eq(Power(a, b), 'a**b')

        # In Python, a**b**c = a**(b**c), whereas Myokit uses a^b^c = (a^b)^c
        self.eq(Power(Power(a, b), c), '(a**b)**c')
        self.eq(Power(a, Power(b, c)), 'a**b**c')

        self.eq(Power(Plus(a, b), c), '(a + b)**c')
        self.eq(Power(a, Minus(b, c)), 'a**(b - c)')
        self.eq(Power(Multiply(a, b), c), '(a * b)**c')
        self.eq(Power(a, Divide(b, c)), 'a**(b / c)')

        self.py(Power(Number(4), Power(Number(2), Number(3))), 65536)
        self.py(Power(Power(Number(2), Number(3)), Number(4)), 4096)

    def test_functions(self):
        a, b = self.ab

        self.eq(Sqrt(a), 'math.sqrt(a)')
        self.eq(Exp(a), 'math.exp(a)')
        self.eq(Log(a), 'math.log(a)')
        self.eq(Log(a, b), 'math.log(a, b)')
        self.eq(Log10(a), 'math.log10(a)')
        self.eq(Sin(a), 'math.sin(a)')
        self.eq(Cos(a), 'math.cos(a)')
        self.eq(Tan(a), 'math.tan(a)')
        self.eq(ASin(a), 'math.asin(a)')
        self.eq(ACos(a), 'math.acos(a)')
        self.eq(ATan(a), 'math.atan(a)')
        self.eq(Floor(a), 'math.floor(a)')
        self.eq(Ceil(a), 'math.ceil(a)')
        self.eq(Abs(a), 'abs(a)')

        self.py(Sqrt(Number(9)), 3)
        self.py(Exp(Number(3)), math.exp(3))
        self.py(Log(Number(3)), math.log(3))
        self.py(Log(Number(8), Number(2)), 3)
        self.py(Log10(Number(100)), 2)
        self.py(Sin(Number(1)), math.sin(1))
        self.py(Cos(Number(1)), math.cos(1))
        self.py(Tan(Number(4)), math.tan(4))
        self.py(ASin(Number(0.4)), math.asin(0.4))
        self.py(ACos(Number(0.4)), math.acos(0.4))
        self.py(ATan(Number(0.4)), math.atan(0.4))
        self.py(Floor(Number(3.9)), 3)
        self.py(Ceil(Number(4.01)), 5)
        self.py(Abs(PrefixMinus(Number(12))), 12)
        self.py(Abs(Number(-13)), 13)

    def test_conditions(self):
        a, b, c, d = self.abcd

        self.eq(Equal(a, b), '(a == b)')
        self.eq(NotEqual(a, b), '(a != b)')
        self.eq(More(b, a), '(b > a)')
        self.eq(Less(d, c), '(d < c)')
        self.eq(MoreEqual(c, a), '(c >= a)')
        self.eq(LessEqual(b, d), '(b <= d)')

        self.eq(And(Equal(a, b), NotEqual(c, d)), '((a == b) and (c != d))')
        self.eq(Or(More(d, c), Less(b, a)), '((d > c) or (b < a))')
        self.eq(Not(Equal(d, d)), '(not (d == d))')

        true = Equal(Number(3), Number(3))
        self.py(true, True)
        false = NotEqual(Number(3), Number(3))
        self.py(false, False)
        self.py(More(Number(5), Number(3)), True)
        self.py(More(Number(3), Number(3)), False)
        self.py(MoreEqual(Number(3), Number(3)), True)
        self.py(Less(Number(3), Number(5)), True)
        self.py(More(Number(3), Number(3)), False)
        self.py(LessEqual(Number(3), Number(3)), True)

        self.py(And(true, false), False)
        self.py(And(true, true), True)
        self.py(Or(true, false), True)
        self.py(Or(true, true), True)
        self.py(Or(false, false), False)
        self.py(Not(true), False)
        self.py(Not(false), True)

    def test_conditionals(self):

        a, b, c, d = self.abcd
        self.eq(If(Equal(a, b), d, c), '(d if (a == b) else c)')
        self.py(If(Equal(Number(1), Number(1)), Number(2), Number(3)), 2)
        self.py(If(Equal(Number(10), Number(1)), Number(2), Number(3)), 3)

        self.eq(Piecewise(NotEqual(d, c), b, a), '(b if (d != c) else a)')
        self.eq(Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4)),
                '(c if (a == b) else (3.0 if (a == d) else 4.0))')

        self.py(Piecewise(Equal(Number(7), Number(7)), Number(2),
                          Equal(Number(7), Number(8)), Number(3),
                          Number(4)), 2)
        self.py(Piecewise(Equal(Number(7), Number(7)), Number(2),
                          Equal(Number(7), Number(7)), Number(3),
                          Number(4)), 2)
        self.py(Piecewise(NotEqual(Number(7), Number(7)), Number(2),
                          Equal(Number(7), Number(7)), Number(3),
                          Number(4)), 3)
        self.py(Piecewise(NotEqual(Number(7), Number(7)), Number(2),
                          NotEqual(Number(8), Number(8)), Number(3),
                          Number(4)), 4)


class NumPyExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """
    Test conversion of expressions to NumPy-compatible Python.
    This is used e.g. to graph expressions.
    Numerical tests are provided.
    """
    _name = 'numpy'
    _target = myokit.formats.python.NumPyExpressionWriter

    def py(self, expression, expected, *arguments):
        """ Test by executing in NumPy, possibly with vector arguments. """
        function = expression.pyfunc(use_numpy=True)
        if arguments:
            arguments = [numpy.array(arg) for arg in arguments]
            values = function(*arguments)
            self.assertEqual(list(values), list(expected))
        else:
            value = function()
            self.assertEqual(value, expected)

    def test_number(self):
        self.eq(Number(2), '2.0')
        self.eq(Number(-1), '-1.0')
        self.eq(Number(14, 'kV'), '14.0')
        self.py(Number(-7, 'mA'), -7)

    def test_name(self):
        self.eq(self.b, 'b')
        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.d), 'COMP.D')

    def test_derivative(self):
        self.eq(myokit.Derivative(self.d), 'dot(d)')

    def test_partial_derivative(self):
        e = myokit.InitialValue(self.a)
        self.assertRaisesRegex(NotImplementedError, 'Initial', self.w.ex, e)

    def test_initial_value(self):
        e = myokit.PartialDerivative(self.a, self.b)
        self.assertRaisesRegex(NotImplementedError, 'Partial', self.w.ex, e)

    def test_prefix_plus(self):
        # Test with numbers
        p = Number(11, 'kV')
        self.eq(PrefixPlus(p), '+11.0')
        self.eq(PrefixPlus(PrefixPlus(p)), '++11.0')
        self.eq(PrefixPlus(Number('+1')), '+1.0')
        self.py(PrefixPlus(Number(3)), 3)
        self.py(PrefixPlus(self.a), [3, 4], [3, 4])

        # Test with operators of precedence SUM, PRODUCT, POWER
        a, b, c = self.abc
        self.eq(PrefixPlus(Plus(a, b)), '+(a + b)')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c), '+(a + b) / c')
        self.eq(PrefixPlus(Remainder(b, a)), '+(b % a)')
        self.eq(Power(PrefixPlus(Power(b, b)), c), '(+b**b)**c')
        self.eq(Power(a, PrefixPlus(Power(b, c))), 'a**(+b**c)')

    def test_prefix_minus(self):
        # Test with numbers
        p = Number(11, 'kV')
        self.eq(PrefixMinus(p), '-11.0')
        self.eq(PrefixMinus(PrefixMinus(p)), '--11.0')
        self.eq(PrefixMinus(Number(-1)), '--1.0')
        self.py(PrefixMinus(Number(3)), -3)
        self.py(PrefixMinus(self.a), [3, -4], [-3, 4])

        # Test with operators of precedence SUM, PRODUCT, POWER
        a, b, c = self.abc
        self.eq(PrefixMinus(Minus(a, b)), '-(a - b)')
        self.eq(Multiply(PrefixMinus(Plus(b, a)), c), '-(b + a) * c')
        self.eq(PrefixMinus(Quotient(b, a)), '-(b // a)')
        self.eq(Power(PrefixMinus(Power(c, b)), a), '(-c**b)**a')
        self.eq(Power(a, PrefixMinus(Power(b, c))), 'a**(-b**c)')

    def test_plus_minus(self):
        a, b, c = self.abc
        self.eq(Plus(a, b), 'a + b')
        self.eq(Plus(Plus(a, b), c), 'a + b + c')
        self.eq(Plus(a, Plus(b, c)), 'a + (b + c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Minus(Minus(a, b), c), 'a - b - c')
        self.eq(Minus(a, Minus(b, c)), 'a - (b - c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Plus(Minus(a, b), c), 'a - b + c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')
        self.eq(Minus(Plus(a, b), c), 'a + b - c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')

        self.py(Plus(Number(4), Number(2)), 6)
        self.py(Minus(Number(5), Number(1.5)), 3.5)
        self.py(Plus(a, Number(2)), [5, 1], [3, -1])
        self.py(Minus(Number(5), b), [2, -5], [3, 10])
        self.py(Plus(a, b), [6, 5, 4], [1, 2, 3], [5, 3, 1])
        self.py(Minus(a, b), [1, 2, 3], [8, 2, 5], [7, 0, 2])

    def test_multiply_divide(self):
        a, b, c = self.abc
        self.eq(Multiply(a, b), 'a * b')
        # Left-to-right, so (a * b) * c is the same as a * b * c...
        self.eq(Multiply(Multiply(a, b), c), 'a * b * c')
        # ...but order-of-operations-wise, a * (b * c) is different!
        self.eq(Multiply(a, Multiply(b, c)), 'a * (b * c)')
        # Note that a user typing a * b * c results in (a * b) * c

        self.eq(Divide(a, b), 'a / b')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')

        self.eq(Divide(Multiply(a, b), c), 'a * b / c')
        self.eq(Multiply(Divide(a, b), c), 'a / b * c')
        self.eq(Divide(a, Multiply(b, c)), 'a / (b * c)')
        self.eq(Multiply(a, Divide(b, c)), 'a * (b / c)')

        self.eq(Remainder(Multiply(a, b), c), 'a * b % c')
        self.eq(Multiply(Quotient(a, b), c), 'a // b * c')
        self.eq(Quotient(a, Multiply(b, c)), 'a // (b * c)')
        self.eq(Multiply(a, Remainder(b, c)), 'a * (b % c)')

        self.eq(Divide(a, Quotient(b, c)), 'a / (b // c)')
        self.eq(Divide(Quotient(b, c), a), 'b // c / a')
        self.eq(Divide(a, Remainder(b, c)), 'a / (b % c)')
        self.eq(Divide(Remainder(b, c), a), 'b % c / a')

        self.eq(Multiply(Minus(a, b), c), '(a - b) * c')
        self.eq(Multiply(a, Plus(b, c)), 'a * (b + c)')
        self.eq(Minus(Multiply(a, b), c), 'a * b - c')
        self.eq(Plus(a, Multiply(b, c)), 'a + b * c')
        self.eq(Divide(Plus(a, b), c), '(a + b) / c')
        self.eq(Divide(a, Minus(b, c)), 'a / (b - c)')
        self.eq(Plus(Divide(a, b), c), 'a / b + c')
        self.eq(Minus(a, Divide(b, c)), 'a - b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')

        self.py(Multiply(Number(7), Number(9)), 63)
        self.py(Divide(Number(5), Number(2)), 2.5)
        self.py(Multiply(a, Number(2)), [6, -2], [3, -1])
        self.py(Divide(a, b), [5, -3, 10], 60, [12, -20, 6])

    def test_power(self):
        a, b, c = self.abc
        self.eq(Power(a, b), 'a**b')

        # In Python, a**b**c = a**(b**c), whereas Myokit uses a^b^c = (a^b)^c
        self.eq(Power(Power(a, b), c), '(a**b)**c')
        self.eq(Power(a, Power(b, c)), 'a**b**c')

        self.eq(Power(Plus(a, b), c), '(a + b)**c')
        self.eq(Power(a, Minus(b, c)), 'a**(b - c)')
        self.eq(Power(Multiply(a, b), c), '(a * b)**c')
        self.eq(Power(a, Divide(b, c)), 'a**(b / c)')

        self.py(Power(Number(4), Power(Number(2), Number(3))), 65536)
        self.py(Power(Power(Number(2), Number(3)), Number(4)), 4096)
        self.py(Power(a, Power(b, c)),
                [65536, 1, 4], [4, 1, 2], [2, 1, 4], [3, 1, 0.5])
        self.py(Power(Power(a, b), c),
                [4096, 0.01], [2, 10.0], [3, 2], [4, -1])

    def test_functions(self):
        a, b, c = self.abc

        self.eq(Sqrt(a), 'numpy.sqrt(a)')
        self.eq(Exp(a), 'numpy.exp(a)')
        self.eq(Log(a), 'numpy.log(a)')
        # Log replaced by divide always adds brackets, because parent element
        # won't have added any (never needed around a log(a, b) function).
        self.eq(Log(a, b), '(numpy.log(a) / numpy.log(b))')
        self.eq(Divide(a, Log(b, c)), 'a / (numpy.log(b) / numpy.log(c))')
        self.eq(Log10(a), 'numpy.log10(a)')
        self.eq(Sin(a), 'numpy.sin(a)')
        self.eq(Cos(a), 'numpy.cos(a)')
        self.eq(Tan(a), 'numpy.tan(a)')
        self.eq(ASin(a), 'numpy.arcsin(a)')
        self.eq(ACos(a), 'numpy.arccos(a)')
        self.eq(ATan(a), 'numpy.arctan(a)')
        self.eq(Floor(a), 'numpy.floor(a)')
        self.eq(Ceil(a), 'numpy.ceil(a)')
        self.eq(Abs(a), 'numpy.abs(a)')

        self.py(Sqrt(a), [3, 4], [9, 16])
        self.py(Exp(a), numpy.exp([2, 3]), [2, 3])
        self.py(Log(a), numpy.log([3, 5]), [3, 5])
        self.py(Log(a, b), [3, 4], [27, 16], [3, 2])
        self.py(Log10(a), [2, 3], [100, 1000])
        self.py(Sin(a), numpy.sin([1, 2]), [1, 2])
        self.py(Cos(a), numpy.cos([3, 4]), [3, 4])
        self.py(Tan(a), numpy.tan([5, 6]), [5, 6])
        self.py(ASin(a), numpy.arcsin([0.1, 0.2]), [0.1, 0.2])
        self.py(ACos(a), numpy.arccos([0.3, 0.4, 0.5]), [0.3, 0.4, 0.5])
        self.py(ATan(a), numpy.arctan([0.1]), [0.1])
        self.py(Floor(a), [3, 3], [3.1, 3.9])
        self.py(Ceil(a), [2, 3], [1.1, 3])
        self.py(Abs(a), [13, 13], [13, -13])
        self.py(Abs(PrefixMinus(a)), [1, 2], [1, -2])

    def test_conditions(self):
        a, b, c, d = self.abcd

        self.eq(Equal(a, b), '(a == b)')
        self.eq(NotEqual(a, b), '(a != b)')
        self.eq(More(b, a), '(b > a)')
        self.eq(Less(d, c), '(d < c)')
        self.eq(MoreEqual(c, a), '(c >= a)')
        self.eq(LessEqual(b, d), '(b <= d)')

        self.eq(And(Equal(a, b), NotEqual(c, d)),
                'numpy.logical_and((a == b), (c != d))')
        self.eq(Or(More(d, c), Less(b, a)),
                'numpy.logical_or((d > c), (b < a))')
        self.eq(Not(Equal(d, d)), 'numpy.logical_not((d == d))')

        true = Equal(Number(3), Number(3))
        self.py(true, True)
        false = NotEqual(Number(3), Number(3))
        self.py(false, False)
        self.py(More(Number(5), Number(3)), True)
        self.py(More(Number(3), Number(3)), False)
        self.py(MoreEqual(Number(3), Number(3)), True)
        self.py(Less(Number(3), Number(5)), True)
        self.py(More(Number(3), Number(3)), False)
        self.py(LessEqual(Number(3), Number(3)), True)
        self.py(And(true, false), False)
        self.py(And(true, true), True)
        self.py(Or(true, false), True)
        self.py(Not(true), False)

    def test_conditionals(self):
        a, b, c, d = self.abcd
        e, f, g = self.efg

        self.eq(If(Equal(a, b), d, c), 'numpy.select([(a == b)], [d], c)')
        self.eq(Piecewise(NotEqual(d, c), b, a),
                'numpy.select([(d != c)], [b], a)')
        self.eq(Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4)),
                'numpy.select([(a == b), (a == d)], [c, 3.0], 4.0)')

        self.py(If(Equal(Number(10), Number(1)), Number(2), Number(3)), 3)
        self.py(If(Equal(a, b), c, d), [3, 2], [2, 2], [2, 3], [3, 4], [1, 2])
        self.py(Piecewise(Equal(a, b), c, Equal(d, e), f, g),
                [1, 2, 30], [1, 1, 1], [1, 2, 3], [1, 1, 1], [1, 4, 5],
                [1, 4, 6], [2, 2, 2], [3, 3, 30])


if __name__ == '__main__':
    unittest.main()



#
# Matlab expression writer and keywords
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.python import PythonExpressionWriter


class MatlabExpressionWriter(PythonExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to a Matlab syntax.
    """
    def __init__(self):
        super().__init__()
        self._function_prefix = ''
        self._fcond = 'ifthenelse'

    def set_condition_function(self, func):
        """
        Sets a function name to output to handle :class:`myokit.If`.

        The function must take arguments ``(condition, value_if_true,
        value_if_false)`` and will be used to handle both :class:`myokit.If`
        and :class:`myokit.Piecewise`.

        By default, `ifthenelse` is used, which the user is expected to define
        if ``If`` or ``Piecewise`` will be used. For example::

            function y = ifthenelse(condition, value_if_true, value_if_false)
                if (condition)
                    y = value_if_true;
                else
                    y = value_if_false;
                end
            end

        """
        if func is not None:
            func = str(func).strip()
        if func is None or func == '':
            raise ValueError(
                'The MatlabExpressionWriter needs a condition function to be'
                ' set.')
        self._fcond = func

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_number(self, e):
    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    def _ex_quotient(self, e):
        # Round towards minus infinity
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])))

    def _ex_remainder(self, e):
        # Uses the round-towards-minus-infinity convention!
        return f'mod({self.ex(e[0])}, {self.ex(e[1])})'

    def _ex_power(self, e):
        # Same associativity as Myokit, not Python! So override
        e1 = f'({self.ex(e[0])})' if e.bracket(e[0]) else f'{self.ex(e[0])}'
        e2 = f'({self.ex(e[1])})' if e.bracket(e[1]) else f'{self.ex(e[1])}'
        return f'{e1}^{e2}'

    #def _ex_sqrt(self, e):
    #    Ignore imaginary part
    #    return 'real(' + self._ex_function(e, 'sqrt') + ')'
    #def _ex_sin(self, e):
    #def _ex_cos(self, e):
    #def _ex_tan(self, e):
    #def _ex_asin(self, e):
    #def _ex_acos(self, e):
    #def _ex_atan(self, e):
    #def _ex_exp(self, e):

    def _ex_log(self, e):
        if len(e) == 1:
            return self._ex_function(e, 'log')
        # Always add brackets: Parent was expecting a function so will never
        # have added them.
        return f'(log({self.ex(e[0])}) / log({self.ex(e[1])}))'

    #def _ex_log10(self, e):
    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):

    def _ex_abs(self, e):
        return self._ex_function(e, 'abs')

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_infix_logical(e, '&&')

    def _ex_or(self, e):
        return self._ex_infix_logical(e, '||')

    def _ex_not(self, e):
        return f'(!{self.ex(e[0])})'

    def _ex_if(self, e):
        _if, _then, _else = self.ex(e._i), self.ex(e._t), self.ex(e._e)
        return f'{self._fcond}({_if}, {_then}, {_else})'

    def _ex_piecewise(self, e):
        # Render ifs; add extra bracket if not a condition (see _ex_if)
        _ifs = [self.ex(x) for x in e._i]
        _thens = [self.ex(x) for x in e._e]

        s = []
        n = len(_ifs)
        for _if, _then in zip(_ifs, _thens):
            s.append(f'{self._fcond}({_if}, {_then}, ')
        s.append(_thens[-1])
        s.append(')' * len(_ifs))
        return ''.join(s)




#
# Stan expression writer and keywords
#
# cell.stan:: This will become the stan model definition file
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.python import PythonExpressionWriter


class StanExpressionWriter(PythonExpressionWriter):
    """
    This:class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit:class:`expressions <myokit.Expression>` to a Stan syntax.
    """
    def __init__(self):
        super().__init__()
        self._function_prefix = ''

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_number(self, e):
    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    def _ex_quotient(self, e):
        # Use floor to round towards minus infinity
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])))

    def _ex_remainder(self, e):
        # Extra brackets needed: Minus has lower precedence than division.
        i = myokit.Minus(e[0], myokit.Multiply(
            e[1], myokit.Floor(myokit.Divide(e[0], e[1]))))
        return f'({self.ex(i)})'

    def _ex_power(self, e):
        # Like Python (and unlike Myokit), Stan uses a right-associative power
        # operator.
        if e.bracket(e[0]) or isinstance(e[0], myokit.Power):
            out = f'({self.ex(e[0])})'
        else:
            out = f'{self.ex(e[0])}'
        if e.bracket(e[1]) and not isinstance(e[1], myokit.Power):
            return f'{out}^({self.ex(e[1])})'
        else:
            return f'{out}^{self.ex(e[1])}'

    #def _ex_sqrt(self, e):
    #def _ex_sin(self, e):
    #def _ex_cos(self, e):
    #def _ex_tan(self, e):
    #def _ex_asin(self, e):
    #def _ex_acos(self, e):
    #def _ex_atan(self, e):
    #def _ex_exp(self, e):

    def _ex_log(self, e):
        if len(e) == 2:
            return f'(log({self.ex(e[0])}) / log({self.ex(e[1])}))'
        return f'log({self.ex(e[0])})'

    def _ex_log10(self, e):
        return f'log10({self.ex(e[0])})'

    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):
    #def _ex_abs(self, e):

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_infix_logical(e, '&&')

    def _ex_or(self, e):
        return self._ex_infix_logical(e, '||')

    def _ex_not(self, e):
        return f'(!{self.ex(e[0])})'

    def _ex_if(self, e):
        return f'({self.ex(e._i)} ? {self.ex(e._t)} : {self.ex(e._e)})'

    def _ex_piecewise(self, e):
        s = []
        for _if, _then in zip(e._i, e._e):
            s.append(f'({self.ex(_if)} ? {self.ex(_then)} : ')
        s.append(self.ex(e._e[-1]))
        s.append(')' * len(e._i))
        return ''.join(s)




#
# Ansi-C expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.python import PythonExpressionWriter


class CBasedExpressionWriter(PythonExpressionWriter):
    """
    Base class for C-style expression writers.
    """
    def __init__(self):
        super().__init__()
        self._function_prefix = ''

    def _ex_prefix(self, e, op):
        # PrefixPlus and PrefixMinus. No simplifications should be made here
        # for PrefixPlus, see https://github.com/myokit/myokit/issues/1054
        x = self.ex(e[0])
        return f'{op}({x})' if e.bracket(e[0]) or x[0] == op else f'{op}{x}'

    def _ex_infix_comparison(self, e, op):
        # For equals etc
        return f'({self.ex(e[0])} {op} {self.ex(e[1])})'

    def _ex_infix_logical(self, e, op):
        # For and and or
        return f'({self.ex(e[0])} {op} {self.ex(e[1])})'

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_initial_value(self, e):
    #def _ex_partial_derivative(self, e):

    #def _ex_number(self, e):

    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):

    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    def _ex_quotient(self, e):
        # Note that this _must_ round towards minus infinity!
        # See myokit.Quotient !
        # No extra brackets needed: Function
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])))

    def _ex_remainder(self, e):
        # Note that this _must_ use the same round-to-neg-inf convention as
        # myokit.Quotient! Implementation below is consistent with Python
        # convention.
        # Extra brackets needed! Minus has lower precedence than division.
        i = myokit.Minus(e[0], myokit.Multiply(
            e[1], myokit.Floor(myokit.Divide(e[0], e[1]))))
        return f'({self.ex(i)})'

    def _ex_power(self, e):
        return f'pow({self.ex(e[0])}, {self.ex(e[1])})'

    #def _ex_sqrt(self, e):
    #def _ex_sin(self, e):
    #def _ex_cos(self, e):
    #def _ex_tan(self, e):
    #def _ex_asin(self, e):
    #def _ex_acos(self, e):
    #def _ex_atan(self, e):
    #def _ex_exp(self, e):

    def _ex_log(self, e):
        if len(e) == 1:
            return self._ex_function(e, 'log')
        # Always add brackets: Parent was expecting a function so will never
        # have added them.
        return f'(log({self.ex(e[0])}) / log({self.ex(e[1])}))'

    #def _ex_log10(self, e):
    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):

    def _ex_abs(self, e):
        return self._ex_function(e, 'fabs')

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_infix_logical(e, '&&')

    def _ex_or(self, e):
        return self._ex_infix_logical(e, '||')

    def _ex_not(self, e):
        # C conditions all have brackets, so don't add more
        return f'(!{self.ex(e[0])})'

    def _ex_if(self, e):
        return f'({self.ex(e._i)} ? {self.ex(e._t)} : {self.ex(e._e)})'

    def _ex_piecewise(self, e):
        # Render ifs; add extra bracket if not a condition (see _ex_if)
        _ifs = [self.ex(x) for x in e._i]
        _thens = [self.ex(x) for x in e._e]

        s = []
        n = len(_ifs)
        for _if, _then in zip(_ifs, _thens):
            s.append(f'({_if} ? {_then} : ')
        s.append(_thens[-1])
        s.append(')' * len(_ifs))
        return ''.join(s)


class AnsiCExpressionWriter(CBasedExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` writes
    equations for variables in a C-style syntax.
    """
    def __init__(self):
        super().__init__()
        self._fcond = None

    def set_condition_function(self, func=None):
        """
        Sets a function name to use for :class:`myokit.If`; if not set the
        ternary operatur will be used.

        If given, the function arguments should be ``(condition, value_if_true,
        value_if_false)``. To revert to using the ternary operator, call with
        ``func=None``.
        """
        self._fcond = func

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):

    def _ex_initial_value(self, e):
        # These are disabled by default, but enabled here to support CVODES
        # sensitivity calculations.
        return self._flhs(e)

    def _ex_partial_derivative(self, e):
        # These are disabled by default, but enabled here to support CVODES
        # sensitivity calculations.
        return self._flhs(e)

    #def _ex_number(self, e):

    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):

    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    #def _ex_quotient(self, e):
    #def _ex_remainder(self, e):
    #def _ex_power(self, e):
    #def _ex_sqrt(self, e):
    #def _ex_sin(self, e):
    #def _ex_cos(self, e):
    #def _ex_tan(self, e):
    #def _ex_asin(self, e):
    #def _ex_acos(self, e):
    #def _ex_atan(self, e):
    #def _ex_exp(self, e):
    #def _ex_log(self, e):
    #def _ex_log10(self, e):
    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):
    #def _ex_abs(self, e):
    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):
    #def _ex_and(self, e):
    #def _ex_or(self, e):
    #def _ex_not(self, e):

    def _ex_if(self, e):
        _if, _then, _else = self.ex(e._i), self.ex(e._t), self.ex(e._e)

        # Use if-then-else function?
        if self._fcond is not None:
            return f'{self._fcond}({_if}, {_then}, {_else})'

        # Default: use ternary operator
        return f'({_if} ? {_then} : {_else})'

    def _ex_piecewise(self, e):
        # Allow _fcond

        # Render ifs; add extra bracket if not a condition (see _ex_if)
        _ifs = [self.ex(x) for x in e._i]
        _thens = [self.ex(x) for x in e._e]

        s = []
        n = len(_ifs)
        if self._fcond is not None:
            for _if, _then in zip(_ifs, _thens):
                s.append(f'{self._fcond}({_if}, {_then}, ')
        else:
            for _if, _then in zip(_ifs, _thens):
                s.append(f'({_if} ? {_then} : ')
        s.append(_thens[-1])
        s.append(')' * len(_ifs))
        return ''.join(s)




#
# CUDA expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.ansic import CBasedExpressionWriter


class CudaExpressionWriter(CBasedExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to their CUDA equivalents.
    """
    def __init__(self, precision=myokit.SINGLE_PRECISION):
        super().__init__()
        self._sp = (precision == myokit.SINGLE_PRECISION)

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_initial_value(self, e):
    #def _ex_partial_derivative(self, e):

    def _ex_number(self, e):
        x = super()._ex_number(e)
        return x + 'f' if self._sp else x

    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    #def _ex_quotient(self, e):
    #def _ex_remainder(self, e):

    def _ex_power(self, e):
        pow = 'powf' if self._sp else 'pow'
        return f'{pow}({self.ex(e[0])}, {self.ex(e[1])})'

    def _ex_sqrt(self, e):
        f = 'sqrtf' if self._sp else 'sqrt'
        return self._ex_function(e, f)

    def _ex_sin(self, e):
        f = 'sinf' if self._sp else 'sin'
        return self._ex_function(e, f)

    def _ex_cos(self, e):
        f = 'cosf' if self._sp else 'cos'
        return self._ex_function(e, f)

    def _ex_tan(self, e):
        f = 'tanf' if self._sp else 'tan'
        return self._ex_function(e, f)

    def _ex_asin(self, e):
        f = 'asinf' if self._sp else 'asin'
        return self._ex_function(e, f)

    def _ex_acos(self, e):
        f = 'acosf' if self._sp else 'acos'
        return self._ex_function(e, f)

    def _ex_atan(self, e):
        f = 'atanf' if self._sp else 'atan'
        return self._ex_function(e, f)

    def _ex_exp(self, e):
        f = 'expf' if self._sp else 'exp'
        return self._ex_function(e, f)

    def _ex_log(self, e):
        f = 'logf' if self._sp else 'log'
        if len(e) == 1:
            return self._ex_function(e, f)
        return '(' + f + '(' + self.ex(e[0]) + ') / ' + f + '(' \
            + self.ex(e[1]) + '))'

    def _ex_log10(self, e):
        f = 'log10f' if self._sp else 'log10'
        return self._ex_function(e, f)

    def _ex_floor(self, e):
        f = 'floorf' if self._sp else 'floor'
        return self._ex_function(e, f)

    def _ex_ceil(self, e):
        f = 'ceilf' if self._sp else 'ceil'
        return self._ex_function(e, f)

    def _ex_abs(self, e):
        f = 'fabsf' if self._sp else 'fabs'
        return self._ex_function(e, f)

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):
    #def _ex_and(self, e):
    #def _ex_or(self, e):
    #def _ex_not(self, e):
    #def _ex_if(self, e):
    #def _ex_piecewise(self, e):




#!/usr/bin/env python3
#
# Tests the expression writer for C++.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats.cpp

import myokit.tests


class CppExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """
    Test the C++ expression writer.
    This inherits from Ansi C, only dropping support for initial values and
    partial derivatives.
    """
    _name = 'cpp'
    _target = myokit.formats.cpp.CppExpressionWriter

    def test_basics(self):
        # Test a few arbitrary expressions, just to check the inheritcance has
        # worked.
        a, b, c, d = self.abcd
        self.eq(myokit.Number(1), '1.0')
        self.eq(myokit.Number(13, 'mV'), '13.0')
        self.eq(myokit.Divide(myokit.PrefixPlus(myokit.Plus(a, b)), c),
                '+(a + b) / c')
        self.eq(myokit.Multiply(myokit.PrefixMinus(myokit.Plus(b, a)), c),
                '-(b + a) * c')
        self.eq(myokit.Multiply(myokit.Divide(a, b), c), 'a / b * c')
        self.eq(myokit.Divide(a, myokit.Multiply(b, c)), 'a / (b * c)')
        self.eq(myokit.Quotient(myokit.Divide(a, c), b), 'floor(a / c / b)')
        self.eq(myokit.ASin(a), 'asin(a)')
        self.eq(myokit.Power(myokit.PrefixMinus(a), b), 'pow(-a, b)')
        self.eq(myokit.Power(a, myokit.Minus(b, c)), 'pow(a, b - c)')
        self.eq(myokit.Power(myokit.Multiply(a, b), c), 'pow(a * b, c)')
        self.eq(myokit.Log(a, b), '(log(a) / log(b))')
        self.eq(myokit.Sin(a), 'sin(a)')
        self.eq(myokit.And(myokit.Equal(a, b), myokit.NotEqual(c, d)),
                '((a == b) && (c != d))')
        self.eq(myokit.Not(myokit.Less(myokit.Number(1), myokit.Number(2))),
                '(!(1.0 < 2.0))')
        self.eq(myokit.If(myokit.Equal(a, b), d, c), '((a == b) ? d : c)')

    def test_derivative(self):
        self.eq(myokit.Derivative(self.a), 'dot(a)')

    def test_partial_derivative(self):
        self.assertRaisesRegex(
            NotImplementedError, 'Partial',
            self.w.ex, myokit.PartialDerivative(self.a, self.b))

    def test_initial_value(self):
        self.assertRaisesRegex(
            NotImplementedError, 'Initial',
            self.w.ex, myokit.InitialValue(self.a))

    def test_conditionals(self):
        # Inherited from AnsiCExpressionWriter

        a, b, c, d = self.abcd
        self.eq(myokit.If(myokit.Equal(a, b), d, c), '((a == b) ? d : c)')
        self.eq(myokit.Piecewise(myokit.NotEqual(d, c), b, a),
                '((d != c) ? b : a)')
        self.eq(myokit.Piecewise(myokit.Equal(a, b), c,
                                 myokit.Equal(a, d), myokit.Number(3),
                                 myokit.Number(4)),
                '((a == b) ? c : ((a == d) ? 3.0 : 4.0))')

        # Using if-then-else function
        w = self._target()
        w.set_lhs_function(lambda v: v.var().name())
        w.set_condition_function('ite')

        self.assertEqual(w.ex(myokit.If(myokit.More(b, a), c, d)),
                         'ite((b > a), c, d)')
        self.assertEqual(w.ex(myokit.Piecewise(myokit.Less(d, c), b, a)),
                         'ite((d < c), b, a)')
        self.assertEqual(
            w.ex(myokit.Piecewise(myokit.Equal(a, b), c,
                                  myokit.Equal(a, d), myokit.Number(3),
                                  myokit.Number(4))),
            'ite((a == b), c, ite((a == d), 3.0, 4.0))')


if __name__ == '__main__':
    unittest.main()



#
# DiffSL expression writer
#
# Supported functions:
#  https://martinjrobins.github.io/diffsl/functions.html
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import warnings

from myokit import And, Equal, If, LessEqual, Log, MoreEqual, Not, Number
from myokit.formats.ansic import CBasedExpressionWriter


class DiffSLExpressionWriter(CBasedExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` writes
    equations for variables in DiffSL syntax.

    For details of the language, see https://martinjrobins.github.io/diffsl/.

    Warnings will be generated if unsupported functions are used in the model.
    Unsupported functions: `acos`, `asin`, `atan`, `ceil`, `floor`.

    Support for logic expressions is implemented with heaviside functions.
    For example, `(a >= b)` is converted to `heaviside(a - b)`.

    """

    def __init__(self):
        super().__init__()

    # -- Literals and identifiers

    # def _ex_name(self, e):
    # def _ex_number(self, e):

    # -- Functions

    def _ex_abs(self, e):
        return self._ex_function(e, 'abs')

    def _ex_acos(self, e):
        warnings.warn('Unsupported function: acos()')
        return super()._ex_acos(e)

    def _ex_asin(self, e):
        warnings.warn('Unsupported function: asin()')
        return super()._ex_asin(e)

    def _ex_atan(self, e):
        warnings.warn('Unsupported function: atan()')
        return super()._ex_atan(e)

    def _ex_ceil(self, e):
        warnings.warn('Unsupported function: ceil()')
        return super()._ex_ceil(e)

    # def _ex_cos(self, e):
    # def _ex_derivative(self, e):
    # def _ex_divide(self, e):
    # def _ex_exp(self, e):

    def _ex_floor(self, e):
        warnings.warn('Unsupported function: floor()')
        return super()._ex_floor(e)

    # def _ex_log(self, e):

    def _ex_log10(self, e):
        # Log10(a) = Log(a, 10.0) -> '(log(a) / log(10.0))'
        return super()._ex_log(Log(e[0], Number(10)))

    # def _ex_minus(self, e):
    # def _ex_multiply(self, e):
    # def _ex_plus(self, e):
    # def _ex_power(self, e):
    # def _ex_prefix_minus(self, e):
    # def _ex_prefix_plus(self, e):
    # def _ex_quotient(self, e):
    # def _ex_remainder(self, e):
    # def _ex_sin(self, e):
    # def _ex_sqrt(self, e):
    # def _ex_tan(self, e):

    # -- Conditional operators

    def _ex_and(self, e):
        # (a and b) == a * b, where a, b are in {0, 1}
        return f'{self.ex(e[0])} * {self.ex(e[1])}'

    def _ex_equal(self, e):
        # (a == b) == heaviside(a - b) * heaviside(b - a)
        return self.ex(And(MoreEqual(e[0], e[1]), LessEqual(e[0], e[1])))

    def _ex_less(self, e):
        # (a < b) == 1 - heaviside(a - b)
        return self.ex(Not(MoreEqual(e[0], e[1])))

    def _ex_less_equal(self, e):
        # (a <= b) == heaviside(b - a)
        return f'heaviside({self.ex(e[1])} - {self.ex(e[0])})'

    def _ex_more(self, e):
        # (a > b) == 1 - heaviside(b - a)
        return self.ex(Not(LessEqual(e[0], e[1])))

    def _ex_more_equal(self, e):
        # (a >= b) == heaviside(a - b)
        return f'heaviside({self.ex(e[0])} - {self.ex(e[1])})'

    def _ex_not(self, e):
        # not(a) == (1 - a), where a is in {0, 1}
        return f'(1 - {self.ex(e[0])})'

    def _ex_not_equal(self, e):
        # (a != b) == 1 - heaviside(a - b) * heaviside(b - a)
        return self.ex(Not(Equal(e[0], e[1])))

    def _ex_or(self, e):
        # a or b == not(not(a) and not(b)), where a, b are in {0, 1}
        return self.ex(Not(And(Not(e[0]), Not(e[1]))))

    # -- Conditional expressions

    def _ex_if(self, e):
        _if = self.ex(e._i)
        _then = self.ex(e._t)
        _not_if = self.ex(Not(e._i))
        _else = self.ex(e._e)

        return f'({_if} * {_then} + {_not_if} * {_else})'

    def _ex_piecewise(self, e):
        # Convert piecewise to nested ifs
        # e.g. piecewise(a, b, c, d, e) -> if(a, b, if(c, d, e))
        n = len(e._i)

        _nested_ifs = e._e[n]

        for i in range(n - 1, -1, -1):
            _nested_ifs = If(e._i[i], e._e[i], _nested_ifs)

        return self._ex_if(_nested_ifs)



#
# Export as a DiffSL model.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections
import warnings

import myokit
import myokit.formats
import myokit.formats.diffsl as diffsl
import myokit.lib.guess as guess
import myokit.lib.markov as markov


class DiffSLExporter(myokit.formats.Exporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates a DiffSL
    implementation of a Myokit model.

    Only the model definition is exported. No inputs are provided, there is
    no protocol defined, and only state variables are output.

    For details of the language, see https://martinjrobins.github.io/diffsl/
    """

    def model(self, path, model, protocol=None, convert_units=True):
        """
        Exports a :class:`myokit.Model` in DiffSL format, writing the result to
        the file indicated by ``path``.

        A :class:`myokit.ExportError` will be raised if any errors occur.
        Warnings will be generated if unsupported functions (e.g. atan) are
        used in the model. For a full list, see
        :class:`myokit.formats.diffsl.DiffSLExpressionWriter
        <DiffSLExpressionWriter>`.

        Arguments:

        ``path``
            The path to write the generated model to.
        ``model``
            The :class:`myokit.Model` to export.
        ``protocol``
            Not implemented!
        ``convert_units``
            If set to ``True`` (default), the method will attempt to convert to
            preferred units for voltage (mV), current (A/F), and time (ms).
        """
        # Raise exception if protocol is set
        if protocol is not None:
            raise ValueError(
                'DiffSL export does not support an input protocol.'
            )

        # Check model validity
        try:
            model.validate()
        except myokit.MyokitError as e:
            raise myokit.ExportError(
                'DiffSL export requires a valid model.'
            ) from e

        # Prepare model for export
        model = self._prep_model(model, convert_units)

        # Generate DiffSL model
        diffsl_model = self._generate_diffsl(model)

        # Write DiffSL model to file
        with open(path, 'w') as f:
            f.write(diffsl_model)

    def _prep_model(self, model, convert_units):
        """
        Prepare the model for export to DiffSL.
        """
        # Rewrite model so that any Markov models have a 1-sum(...) state
        # This also clones the model, so that changes can be made
        model = markov.convert_markov_models_to_compact_form(model)

        # Remove all model bindings apart from time and pace
        _ = myokit._prepare_bindings(
            model,
            {
                'time': 't',
                'pace': 'pace',
            },
        )

        # Remove hardcoded stimulus protocol, if any
        guess.remove_embedded_protocol(model)

        # Get / try to guess some model variables
        time = model.time()  # engine.time
        vm = guess.membrane_potential(model)  # Vm
        cm = guess.membrane_capacitance(model)  # Cm
        currents = self._guess_currents(model)

        # Check for explicit time dependence
        time_refs = list(time.refs_by())
        if time_refs:
            raise myokit.ExportError(
                'DiffSL export does not support explicit time dependence:\n'
                + '\n'.join([f'{v} = {v.rhs()}' for v in time_refs])
            )

        if convert_units:
            # Convert currents to A/F
            helpers = [] if cm is None else [cm.rhs()]
            for var in currents:
                self._convert_current_unit(var, helpers=helpers)

            # Convert potentials to mV
            self._convert_potential_unit(vm)

            # Convert time to ms
            if time.unit() != myokit.units.ms:
                self._convert_unit(time, 'ms')

        # Add intermediary variables for state derivatives with rhs references
        # Before:
        #   dot(x) = x / 5
        #   y = 1 + dot(x)
        # After:
        #   dot_x =  x / 5
        #   dot(x) = dot_x
        #   y = 1 + dot_x
        model.remove_derivative_references()

        return model

    def _generate_diffsl(self, model):
        """
        Generate a DiffSL model from a prepped Myokit model.
        DiffSL inputs will be left empty, and outputs will be set to
        state variables in alphabetical order.
        """

        # Create DiffSL-compatible variable names
        var_to_name = self._create_diffsl_variable_names(model)

        # Create a naming function
        def var_name(e):
            if isinstance(e, myokit.Derivative):
                return var_to_name[e]
            elif isinstance(e, myokit.LhsExpression):
                return var_to_name[e.var()]
            elif isinstance(e, myokit.Variable):
                return var_to_name[e]
            raise ValueError(  # pragma: no cover
                'Not a variable or LhsExpression: ' + str(e)
            )

        # Create an expression writer
        e = diffsl.DiffSLExpressionWriter()
        e.set_lhs_function(var_name)

        export_lines = []  # DiffSL export lines
        tab = '  '  # Tab character

        # Sort equations in solvable order (grouped by component)
        sorted_eqs = model.solvable_order()

        # Variables to be excluded from output or handled separately.
        # State derivatives are handled in dudt_i, F_i and G_i; time is
        # excluded from the output; pace is handled separately.
        time = model.time()
        pace = model.binding('pace')
        special_vars = set(v for v in model.states())
        special_vars.add(time)
        if pace is not None:
            special_vars.add(pace)

        # Add metadata
        export_lines.append('/*')
        export_lines.append('This file was generated by Myokit.')
        if model.meta:
            export_lines.append('')
            for key, val in sorted(model.meta.items()):
                export_lines.append(f'{key}: {val}')
            export_lines.append('')
        export_lines.append('*/')
        export_lines.append('')

        # Add empty input parameter list
        export_lines.append('/* Input parameters */')
        export_lines.append('/* E.g. in = [ varZero, varOne, varTwo ] */')
        export_lines.append('in = [ ]')
        export_lines.append('')

        # Add pace
        if pace is not None:
            export_lines.append('/* Engine: pace */')
            export_lines.append('/* E.g.')
            export_lines.append('  -80 * (1 - sigmoid((t-100)*5000))')
            export_lines.append(
                '  -120 * (sigmoid((t-100)*5000) - sigmoid((t-200)*5000))'
            )
            export_lines.append('*/')

            lhs = var_name(pace)
            rhs = e.ex(pace.rhs())
            qname = pace.qname()
            unit = '' if pace.unit() is None else f' {pace.unit()}'
            export_lines.append(f'{lhs} {{ {rhs} }} /* {qname}{unit} */')
            export_lines.append('')

        # Add constants
        const_vars = (
            set(model.variables(const=True, deep=True, state=False))
            - special_vars
        )
        for component_label, eq_list in sorted_eqs.items():
            const_eqs = [
                eq for eq in eq_list.equations() if eq.lhs.var() in const_vars
            ]
            if const_eqs:
                export_lines.append(f'/* Constants: {component_label} */')
                for eq in const_eqs:
                    v = eq.lhs.var()
                    lhs = var_name(v)
                    rhs = e.ex(eq.rhs)
                    qname = v.qname()
                    unit = '' if v.unit() is None else f' {v.unit()}'
                    export_lines.append(
                        f'{lhs} {{ {rhs} }} /* {qname}{unit} */'
                    )
                export_lines.append('')

        # Add initial conditions `u_i`
        export_lines.append('/* Initial conditions */')
        export_lines.append('u_i {')
        for v in model.states():
            lhs = var_name(v)
            rhs = v.initial_value()
            qname = v.qname()
            unit = '' if v.unit() is None else f' {v.unit()}'
            export_lines.append(f'{tab}{lhs} = {rhs}, /* {qname}{unit} */')
        export_lines.append('}')
        export_lines.append('')

        # Add state derivatives `dudt_i`
        export_lines.append('dudt_i {')
        for v in model.states():
            lhs = var_name(v.lhs())
            export_lines.append(f'{tab}{lhs} = 0,')
        export_lines.append('}')
        export_lines.append('')

        # Add remaining variables
        todo_vars = (
            set(model.variables(const=False, deep=True, state=False))
            - special_vars
        )
        for component_label, eq_list in sorted_eqs.items():
            todo_eqs = [
                eq for eq in eq_list.equations() if eq.lhs.var() in todo_vars
            ]
            if todo_eqs:
                export_lines.append(f'/* Variables: {component_label} */')
                for eq in todo_eqs:
                    v = eq.lhs.var()
                    lhs = var_name(v)
                    rhs = e.ex(eq.rhs)
                    qname = v.qname()
                    unit = '' if v.unit() is None else f' {v.unit()}'
                    export_lines.append(
                        f'{lhs} {{ {rhs} }} /* {qname}{unit} */'
                    )
                export_lines.append('')

        # Add `F_i`
        export_lines.append('/* Solve */')
        export_lines.append('F_i {')
        for v in model.states():
            lhs = var_name(v.lhs())
            export_lines.append(f'{tab}{lhs},')
        export_lines.append('}')
        export_lines.append('')

        # Add `G_i`
        export_lines.append('G_i {')
        for v in model.states():
            rhs = e.ex(v.rhs())
            export_lines.append(f'{tab}{rhs},')
        export_lines.append('}')
        export_lines.append('')

        # Output state variables in alphabetical order
        export_lines.append('/* Output */')
        export_lines.append('out_i {')
        vars = list(model.states())
        for v in sorted(vars, key=lambda x: var_name(x).swapcase()):
            lhs = var_name(v)
            export_lines.append(f'{tab}{lhs},')
        export_lines.append('}')
        export_lines.append('')

        return '\n'.join(export_lines)

    def _convert_current_unit(self, var, helpers=None):
        """
        Convert a current to A/F if its present unit isn't recommended.
        """
        recommended_units = [
            myokit.parse_unit('pA/pF'),
            myokit.parse_unit('uA/cm^2'),
        ]

        if var.unit() not in recommended_units:
            self._convert_unit(var, 'A/F', helpers=helpers)

    def _convert_potential_unit(self, var, helpers=None):
        """
        Convert a potential to mV if its present unit isn't recommended.
        """
        recommended_units = [myokit.units.mV]

        if var.unit() not in recommended_units:
            self._convert_unit(var, 'mV', helpers=helpers)

    def _convert_unit(self, var, unit, helpers=None):
        """
        Convert a variable to the given unit if possible. Throws a warning if
        the conversion is not possible.
        """
        if var.unit() is None:
            return

        try:
            var.convert_unit(unit, helpers=helpers)
        except myokit.IncompatibleUnitError:
            warnings.warn(
                'Unable to convert ' + var.qname() + ' to recommended'
                ' units of ' + str(unit) + '.'
            )

    def _create_diffsl_variable_names(self, model):
        """
        Create DiffSL-compatible names for all variables in the model.

        The following strategy is followed:
         - Fully qualified names are used for all variables.
         - Variables are checked for special names, and changed if necessary.
         - Unsupported characters like '.' and '_' are replaced.
         - Any conflicts are resolved in a final step by appending a number.
        """

        # Convert name to a DiffSL-compatible variable name
        def convert_name(name):
            # Remove unsupported chars like '_' and '.', and stagger case.
            # Preserves existing staggered case in names e.g.
            # voltage_clamp.R_seal_MOhm -> voltageClampRSealMOhm
            # voltageClamp.RSealMOhm -> voltageClampRSealMOhm
            name_chars = []
            caps_flag = False
            for ch in name:
                if ch.isalpha():
                    if caps_flag:
                        name_chars.append(ch.upper())
                        caps_flag = False
                    else:
                        name_chars.append(ch)
                elif ch.isdigit():
                    name_chars.append(ch)
                    caps_flag = True
                else:
                    caps_flag = True

            return ''.join(name_chars)

        var_to_name = collections.OrderedDict()

        # Store initial names for variables
        for var in model.variables(deep=True, sort=True):
            var_to_name[var] = convert_name(var.qname())
            if var.is_state():
                var_to_name[var.lhs()] = convert_name('diff_' + var.qname())

        # Check for conflicts with known keywords
        from . import keywords

        needs_renaming = collections.OrderedDict()
        for keyword in keywords:
            needs_renaming[keyword] = []

        # Find naming conflicts, create inverse mapping
        name_to_var = collections.OrderedDict()
        for var, name in var_to_name.items():

            # Known conflict?
            if name in needs_renaming:
                needs_renaming[name].append(var)
                continue

            # Test for new conflicts
            var2 = name_to_var.get(name, None)
            if var2 is not None:
                needs_renaming[name] = [var2, var]
                continue

            name_to_var[name] = var

        # Resolve naming conflicts
        for name, variables in needs_renaming.items():
            # Add a number to the end of the name, increasing until unique
            i = 1
            root = name
            for var in variables:
                name = f'{root}{i}'
                while name in name_to_var:
                    i += 1
                    name = f'{root}{i}'
                var_to_name[var] = name
                name_to_var[name] = var

        return var_to_name

    def _guess_currents(self, model):
        """
        Tries to make a list of membrane currents. Removes potentials
        from the guessed list.
        """
        unmatch_units = [
            myokit.units.mV,
            myokit.units.V,
        ]
        guessed_currents = guess.membrane_currents(model)
        return [x for x in guessed_currents if x.unit() not in unmatch_units]

    def supports_model(self):
        """See :meth:`myokit.formats.Exporter.supports_model()`."""
        return True



#!/usr/bin/env python3
#
# Tests the DiffSL module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import itertools
import unittest

import myokit
import myokit.formats
import myokit.formats.diffsl
import myokit.tests
from myokit import (Abs, ACos, And, ASin, ATan, Ceil, Cos, Divide, Equal, Exp,
                    Floor, If, Less, LessEqual, Log, Log10, Minus, More,
                    MoreEqual, Multiply, Not, NotEqual, Number, Or, Piecewise,
                    Plus, Power, PrefixMinus, PrefixPlus, Quotient, Remainder,
                    Sin, Sqrt, Tan)
from myokit.tests import TemporaryDirectory, WarningCollector

# Model that requires unit conversion
units_model = """
[[model]]
mb.V = -0.08
hh.x = 0.1
hh.y = 0.9
mm.C = 0.9

[engine]
pace = 0 bind pace
time = 0 [s]
    in [s]
    bind time

[ideal]
Vc = engine.pace * 1

[mb]
dot(V) = (hh.I1 + mm.I2 + tt.I3) / C
    in [V]
C = 20 [pF]
    in [pF]

[hh]
dot(x) = (inf - x) / tau
    inf = 0.8
    tau = 3 [s]
        in [s]
dot(y) = alpha * (1 - y) - beta * y
    alpha = 0.1 [1/s]
        in [1/s]
    beta = 0.2 [1/s]
        in [1/s]
I1 = 3 [pS] * x * y * (mb.V - 0.05 [V])
    in [pA]

[mm]
dot_C = beta * O - alpha * C
dot(C) = dot_C
alpha = 0.3 [1/s]
    in [1/s]
beta = 0.4 [1/s]
    in [1/s]
O = 1 - C
I2 = 2 [pS] * O * (mb.V + 0.02 [V])
    in [pA]

[tt]
I3 = 4 [pS] * mm.O * (mb.V + 0.02 [V])
tTest = 0
t_Test = 1
t_test = 2
"""

units_output = """
/*
This file was generated by Myokit.
*/

/* Input parameters */
/* E.g. in = [ varZero, varOne, varTwo ] */
in = [ ]

/* Engine: pace */
/* E.g.
  -80 * (1 - sigmoid((t-100)*5000))
  -120 * (sigmoid((t-100)*5000) - sigmoid((t-200)*5000))
*/
enginePace { 0.0 } /* engine.pace */

/* Constants: hh */
hhXInf { 0.8 } /* hh.x.inf */
hhXTau { 3.0 } /* hh.x.tau [s] */
hhYAlpha { 0.1 } /* hh.y.alpha [S/F] */
hhYBeta { 0.2 } /* hh.y.beta [S/F] */

/* Constants: mm */
mmAlpha { 0.3 } /* mm.alpha [S/F] */
mmBeta { 0.4 } /* mm.beta [S/F] */

/* Constants: tt */
ttTTest1 { 0.0 } /* tt.tTest */
ttTTest2 { 1.0 } /* tt.t_Test */
ttTTest3 { 2.0 } /* tt.t_test */

/* Constants: mb */
mbC { 20.0 } /* mb.C [pF] */

/* Initial conditions */
u_i {
  mbV = -0.08 * 1000 [1 (0.001)], /* mb.V [mV] */
  hhX = 0.1, /* hh.x */
  hhY = 0.9, /* hh.y */
  mmC = 0.9, /* mm.C */
}

dudt_i {
  diffMbV = 0,
  diffHhX = 0,
  diffHhY = 0,
  diffMmC = 0,
}

/* Variables: hh */
hhI1 { 3.0 * hhX * hhY * (mbV / 1000.0 - 0.05) * 0.05 } /* hh.I1 [A/F] */

/* Variables: mm */
mmO { 1.0 - mmC } /* mm.O */
mmI2 { 2.0 * mmO * (mbV / 1000.0 + 0.02) * 0.05 } /* mm.I2 [A/F] */
mmDotC { mmBeta * mmO - mmAlpha * mmC } /* mm.dot_C */

/* Variables: ideal */
idealVc { enginePace * 1.0 } /* ideal.Vc */

/* Variables: tt */
ttI3 { 4.0 * mmO * (mbV / 1000.0 + 0.02) } /* tt.I3 */

/* Solve */
F_i {
  diffMbV,
  diffHhX,
  diffHhY,
  diffMmC,
}

G_i {
  (hhI1 / 0.05 + mmI2 / 0.05 + ttI3) / mbC * 1000.0 / 1000.0,
  (hhXInf - hhX) / hhXTau / 1000.0,
  (hhYAlpha * (1.0 - hhY) - hhYBeta * hhY) / 1000.0,
  mmDotC / 1000.0,
}

/* Output */
out_i {
  hhX,
  hhY,
  mbV,
  mmC,
}
"""


class DiffSLExporterTest(unittest.TestCase):
    """Tests DiffSL export."""

    def test_diffsl_exporter(self):
        # Tests exporting a model

        model = myokit.load_model('example')
        with TemporaryDirectory() as d:
            path = d.path('diffsl.model')

            e = myokit.formats.diffsl.DiffSLExporter()

            # Test with simple model
            e.model(path, model)

            # Test with protocol set
            with self.assertRaisesRegex(ValueError, 'input protocol'):
                e.model(path, model, protocol='-80 + 120*heaviside(t-10)')

            # Test with extra bound variables
            model.get('membrane.C').set_binding('hello')
            e.model(path, model)

            # Test without V being a state variable
            v = model.get('membrane.V')
            v.demote()
            v.set_rhs(3)
            e.model(path, model)

            # Test with explicit time dependence
            t0 = model.get('membrane').add_variable('t0')
            t0.set_rhs('0 + engine.time')
            with self.assertRaisesRegex(myokit.ExportError, 'time dependence'):
                e.model(path, model)

            # Test with invalid model
            v.set_rhs('2 * V')
            with self.assertRaisesRegex(myokit.ExportError, 'valid model'):
                e.model(path, model)

    def test_unit_conversion(self):
        # Tests exporting a model that requires unit conversion

        # Export model
        m = myokit.parse_model(units_model)
        e = myokit.formats.diffsl.DiffSLExporter()
        with TemporaryDirectory() as d:
            path = d.path('diffsl.model')
            e.model(path, m)
            with open(path, 'r') as f:
                observed = f.read().strip().splitlines()

        # Get expected output
        expected = units_output.strip().splitlines()

        # Compare (line by line, for readable output)
        for ob, ex in zip(observed, expected):
            self.assertEqual(ob, ex)
        self.assertEqual(len(observed), len(expected))

        # Test warnings are raised if conversion fails
        m.get('mb.V').set_rhs('hh.I1 + mm.I2')
        m.get('mb').remove_variable(m.get('mb.C'))
        with TemporaryDirectory() as d:
            path = d.path('diffsl.model')
            with WarningCollector() as c:
                e.model(path, m)
            self.assertIn('Unable to convert hh.I1', c.text())
            self.assertIn('Unable to convert mm.I2', c.text())

        m.get('engine.time').set_unit(myokit.units.cm)
        with TemporaryDirectory() as d:
            path = d.path('diffsl.model')
            with WarningCollector() as c:
                e.model(path, m)
            self.assertIn('Unable to convert engine.time', c.text())

    def test_diffsl_exporter_fetching(self):
        # Tests getting an DiffSL exporter via the 'exporter' interface

        e = myokit.formats.exporter('diffsl')
        self.assertIsInstance(e, myokit.formats.diffsl.DiffSLExporter)

    def test_capability_reporting(self):
        # Tests if the correct capabilities are reported
        e = myokit.formats.diffsl.DiffSLExporter()
        self.assertTrue(e.supports_model())


class DiffSLExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """Test conversion to DiffSL syntax."""

    _name = 'diffsl'
    _target = myokit.formats.diffsl.DiffSLExpressionWriter

    def test_number(self):
        self.eq(Number(1), '1.0')
        self.eq(Number(-2), '-2.0')
        self.eq(Number(13, 'mV'), '13.0')

    def test_name(self):
        # Inherited from CBasedExpressionWriter
        self.eq(self.a, 'a')
        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.a), 'COMP.A')

    def test_derivative(self):
        # Inherited from CBasedExpressionWriter
        self.eq(myokit.Derivative(self.a), 'dot(a)')

    def test_partial_derivative(self):
        e = myokit.PartialDerivative(self.a, self.b)
        self.assertRaisesRegex(NotImplementedError, 'Partial', self.w.ex, e)

    def test_initial_value(self):
        e = myokit.InitialValue(self.a)
        self.assertRaisesRegex(NotImplementedError, 'Initial', self.w.ex, e)

    def test_prefix_plus_minus(self):
        # Inherited from CBasedExpressionWriter
        p = Number(11, 'kV')
        a, b, c = self.abc
        self.eq(PrefixPlus(p), '+11.0')
        self.eq(PrefixPlus(PrefixPlus(PrefixPlus(p))), '+(+(+11.0))')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c), '+(a + b) / c')
        self.eq(PrefixMinus(p), '-11.0')
        self.eq(PrefixMinus(PrefixMinus(p)), '-(-11.0)')
        self.eq(PrefixMinus(Number(-1)), '-(-1.0)')
        self.eq(PrefixMinus(Minus(a, b)), '-(a - b)')
        self.eq(Multiply(PrefixMinus(Plus(b, a)), c), '-(b + a) * c')
        self.eq(PrefixMinus(Divide(b, a)), '-(b / a)')

    def test_plus_minus(self):
        a, b, c = self.abc
        self.eq(Plus(a, b), 'a + b')
        self.eq(Plus(Plus(a, b), c), 'a + b + c')
        self.eq(Plus(a, Plus(b, c)), 'a + (b + c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Minus(Minus(a, b), c), 'a - b - c')
        self.eq(Minus(a, Minus(b, c)), 'a - (b - c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Plus(Minus(a, b), c), 'a - b + c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')
        self.eq(Minus(Plus(a, b), c), 'a + b - c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')

        # No expm in DiffSL
        self.eq(Minus(Exp(Number(2)), Number(1)), 'exp(2.0) - 1.0')
        self.eq(Minus(Number(1), Exp(Number(3))), '1.0 - exp(3.0)')

    def test_multiply_divide(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        self.eq(Multiply(a, b), 'a * b')
        self.eq(Multiply(Multiply(a, b), c), 'a * b * c')
        self.eq(Multiply(a, Multiply(b, c)), 'a * (b * c)')
        self.eq(Divide(a, b), 'a / b')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')

    def test_quotient(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        with WarningCollector():
            self.eq(Quotient(a, b), 'floor(a / b)')
            self.eq(Quotient(Plus(a, c), b), 'floor((a + c) / b)')
            self.eq(Quotient(Divide(a, c), b), 'floor(a / c / b)')
            self.eq(Quotient(a, Divide(b, c)), 'floor(a / (b / c))')
            self.eq(Multiply(Quotient(a, b), c), 'floor(a / b) * c')
            self.eq(Multiply(c, Quotient(a, b)), 'c * (floor(a / b))')

    def test_remainder(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        with WarningCollector():
            self.eq(Remainder(a, b), '(a - b * floor(a / b))')
            self.eq(
                Remainder(Plus(a, c), b), '(a + c - b * floor((a + c) / b))'
            )
            self.eq(Multiply(Remainder(a, b), c), '(a - b * floor(a / b)) * c')
            self.eq(Divide(c, Remainder(b, a)), 'c / ((b - a * floor(b / a)))')

    def test_power(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        self.eq(Power(a, b), 'pow(a, b)')
        self.eq(Power(Power(a, b), c), 'pow(pow(a, b), c)')
        self.eq(Power(a, Power(b, c)), 'pow(a, pow(b, c))')

    def test_log(self):
        # Inherited from CBasedExpressionWriter
        a, b = self.ab
        self.eq(Log(a), 'log(a)')
        self.eq(Log10(a), '(log(a) / log(10.0))')
        self.eq(Log(a, b), '(log(a) / log(b))')

    def test_supported_functions(self):
        a = self.a

        self.eq(Abs(a), 'abs(a)')
        self.eq(Cos(a), 'cos(a)')
        self.eq(Exp(a), 'exp(a)')
        self.eq(Log(a), 'log(a)')
        self.eq(Sin(a), 'sin(a)')
        self.eq(Sqrt(a), 'sqrt(a)')
        self.eq(Tan(a), 'tan(a)')

    def test_unsupported_functions(self):
        a = self.a

        with WarningCollector() as wc:
            self.eq(ACos(a), 'acos(a)')
        self.assertIn('Unsupported', wc.text())

        with WarningCollector() as wc:
            self.eq(ASin(a), 'asin(a)')
        self.assertIn('Unsupported', wc.text())

        with WarningCollector() as wc:
            self.eq(ATan(a), 'atan(a)')
        self.assertIn('Unsupported', wc.text())

        with WarningCollector() as wc:
            self.eq(Ceil(a), 'ceil(a)')
        self.assertIn('Unsupported', wc.text())

        with WarningCollector() as wc:
            self.eq(Floor(a), 'floor(a)')
        self.assertIn('Unsupported', wc.text())

    def test_conditional_operators(self):
        a, b, c, d = self.abcd

        self.eq(Equal(a, b), 'heaviside(a - b) * heaviside(b - a)')

        self.eq(Less(a, b), '(1 - heaviside(a - b))')

        self.eq(LessEqual(a, b), 'heaviside(b - a)')

        self.eq(More(a, b), '(1 - heaviside(b - a))')

        self.eq(MoreEqual(a, b), 'heaviside(a - b)')

        self.eq(NotEqual(a, b), '(1 - heaviside(a - b) * heaviside(b - a))')

        self.eq(
            Not(NotEqual(a, b)),
            '(1 - (1 - heaviside(a - b) * heaviside(b - a)))',
        )

        self.eq(
            Not(Not(Equal(a, b))),
            '(1 - (1 - heaviside(a - b) * heaviside(b - a)))',
        )

        self.eq(
            And(Equal(a, b), NotEqual(c, d)),
            'heaviside(a - b) * heaviside(b - a)'
            ' * (1 - heaviside(c - d) * heaviside(d - c))',
        )

        self.eq(
            Or(More(d, c), MoreEqual(b, a)),
            '(1 - (1 - (1 - heaviside(c - d))) * (1 - heaviside(b - a)))',
        )

        self.eq(
            Or(Less(d, c), LessEqual(b, a)),
            '(1 - (1 - (1 - heaviside(d - c))) * (1 - heaviside(a - b)))',
        )

        self.eq(
            Not(Or(Equal(Number(1), Number(2)), Equal(Number(3), Number(4)))),
            '(1 - (1 - (1 - heaviside(1.0 - 2.0) * heaviside(2.0 - 1.0)) '
            '* (1 - heaviside(3.0 - 4.0) * heaviside(4.0 - 3.0))))',
        )

        self.eq(
            Not(Less(Number(1), Number(2))), '(1 - (1 - heaviside(1.0 - 2.0)))'
        )

    def test_if_expressions(self):
        a, b, c, d = self.abcd

        self.eq(
            If(Equal(a, b), c, d),
            '(heaviside(a - b) * heaviside(b - a) * c'
            ' + (1 - heaviside(a - b) * heaviside(b - a)) * d)',
        )

        self.eq(
            If(Equal(a, b), c, Number(0)),
            '(heaviside(a - b) * heaviside(b - a) * c'
            ' + (1 - heaviside(a - b) * heaviside(b - a)) * 0.0)',
        )

        self.eq(
            If(Equal(a, b), Number(0), d),
            '(heaviside(a - b) * heaviside(b - a) * 0.0'
            ' + (1 - heaviside(a - b) * heaviside(b - a)) * d)',
        )

        self.eq(
            If(Equal(a, b), c, Number(1)),
            '(heaviside(a - b) * heaviside(b - a) * c'
            ' + (1 - heaviside(a - b) * heaviside(b - a)) * 1.0)',
        )

        self.eq(
            If(Equal(a, b), Number(1), d),
            '(heaviside(a - b) * heaviside(b - a) * 1.0'
            ' + (1 - heaviside(a - b) * heaviside(b - a)) * d)',
        )

        self.eq(
            If(NotEqual(a, b), c, d),
            '((1 - heaviside(a - b) * heaviside(b - a)) * c'
            ' + (1 - (1 - heaviside(a - b) * heaviside(b - a))) * d)',
        )

        self.eq(
            If(More(a, b), c, d),
            '((1 - heaviside(b - a)) * c + (1 - (1 - heaviside(b - a))) * d)',
        )

        self.eq(
            If(MoreEqual(a, b), c, d),
            '(heaviside(a - b) * c + (1 - heaviside(a - b)) * d)',
        )

        self.eq(
            If(Less(a, b), c, d),
            '((1 - heaviside(a - b)) * c + (1 - (1 - heaviside(a - b))) * d)',
        )

        self.eq(
            If(LessEqual(a, b), c, d),
            '(heaviside(b - a) * c + (1 - heaviside(b - a)) * d)',
        )

    def test_piecewise_expressions(self):
        a, b, c, d = self.abcd

        self.eq(Piecewise(Equal(a, b), c, d), self.w.ex(If(Equal(a, b), c, d)))

        self.eq(
            Piecewise(NotEqual(a, b), c, d),
            self.w.ex(
                If(NotEqual(a, b), c, d),
            ),
        )

        self.eq(
            Piecewise(More(a, b), c, d),
            self.w.ex(
                If(More(a, b), c, d),
            ),
        )

        self.eq(
            Piecewise(MoreEqual(a, b), c, d),
            self.w.ex(If(MoreEqual(a, b), c, d)),
        )

        self.eq(Piecewise(Less(a, b), c, d), self.w.ex(If(Less(a, b), c, d)))

        self.eq(
            Piecewise(LessEqual(a, b), c, d),
            self.w.ex(
                If(LessEqual(a, b), c, d),
            ),
        )

        self.eq(
            Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4)),
            self.w.ex(
                If(Equal(a, b), c, If(Equal(a, d), Number(3), Number(4)))
            ),
        )

        self.eq(
            Piecewise(Less(a, b), Number(0), Less(c, d), Number(0), Number(5)),
            '((1 - heaviside(a - b)) * 0.0 '
            '+ (1 - (1 - heaviside(a - b))) * ((1 - heaviside(c - d)) * 0.0 '
            '+ (1 - (1 - heaviside(c - d))) * 5.0))',
        )

    def test_heaviside_numerical(self):
        """Test generated heaviside expressions with numerical values"""

        def heaviside(x):
            return 1 if x >= 0 else 0

        values = itertools.product(
            [-10e9, -1, -1e-9, 0, 1e-9, 1, 10e9], repeat=4
        )

        for a, b, c, d in values:
            # a == b
            result = int(a == b)
            expr = self.w.ex(Equal(Number(a), Number(b)))
            self.assertEqual(eval(expr), result)

            # a < b
            result = int(a < b)
            expr = self.w.ex(Less(Number(a), Number(b)))
            self.assertEqual(eval(expr), result)

            # a <= b
            result = int(a <= b)
            expr = self.w.ex(LessEqual(Number(a), Number(b)))
            self.assertEqual(eval(expr), result)

            # a > b
            result = int(a > b)
            expr = self.w.ex(More(Number(a), Number(b)))
            self.assertEqual(eval(expr), result)

            # a >= b
            result = int(a >= b)
            expr = self.w.ex(MoreEqual(Number(a), Number(b)))
            self.assertEqual(eval(expr), result)

            # a != b
            result = int(a != b)
            expr = self.w.ex(NotEqual(Number(a), Number(b)))
            self.assertEqual(eval(expr), result)

            # not(a != b)
            result = int(not (a != b))
            expr = self.w.ex(Not(NotEqual(Number(a), Number(b))))
            self.assertEqual(eval(expr), result)

            # not(not(a == b))
            result = int(not (not (a == b)))
            expr = self.w.ex(Not(Not(Equal(Number(a), Number(b)))))
            self.assertEqual(eval(expr), result)

            # (a == b) and (c != d)
            result = int((a == b) and (c != d))
            expr = self.w.ex(
                And(
                    Equal(Number(a), Number(b)), NotEqual(Number(c), Number(d))
                )
            )
            self.assertEqual(eval(expr), result)

            # (d > c) or (b >= a)
            result = int((d > c) or (b >= a))
            expr = self.w.ex(
                Or(More(Number(d), Number(c)), MoreEqual(Number(b), Number(a)))
            )
            self.assertEqual(eval(expr), result)

            # (d < c) or (b <= a)
            result = int((d < c) or (b <= a))
            expr = self.w.ex(
                Or(Less(Number(d), Number(c)), LessEqual(Number(b), Number(a)))
            )
            self.assertEqual(eval(expr), result)

            # (a == b) or (c == d)
            result = int((a == b) or (c == d))
            expr = self.w.ex(
                Or(Equal(Number(a), Number(b)), Equal(Number(c), Number(d)))
            )
            self.assertEqual(eval(expr), result)

            # not(a < b)
            result = int(not (a < b))
            expr = self.w.ex(Not(Less(Number(a), Number(b))))
            self.assertEqual(eval(expr), result)

            # if(a > b, c, d)
            result = c if (a > b) else d
            expr = self.w.ex(
                If(More(Number(a), Number(b)), Number(c), Number(d))
            )
            self.assertEqual(eval(expr), result)

            # if(a >= b, c, d)
            result = c if (a >= b) else d
            expr = self.w.ex(
                If(MoreEqual(Number(a), Number(b)), Number(c), Number(d))
            )
            self.assertEqual(eval(expr), result)

            # if(a < b, c, d)
            result = c if (a < b) else d
            expr = self.w.ex(
                If(Less(Number(a), Number(b)), Number(c), Number(d))
            )
            self.assertEqual(eval(expr), result)

            # if(a <= b, c, d)
            result = c if (a <= b) else d
            expr = self.w.ex(
                If(LessEqual(Number(a), Number(b)), Number(c), Number(d))
            )
            self.assertEqual(eval(expr), result)

            # piecewise(a > b, c, d)
            result = c if (a > b) else d
            expr = self.w.ex(
                Piecewise(More(Number(a), Number(b)), Number(c), Number(d))
            )
            self.assertEqual(eval(expr), result)

            # piecewise(a >= b, c, d)
            result = c if (a >= b) else d
            expr = self.w.ex(
                Piecewise(
                    MoreEqual(Number(a), Number(b)), Number(c), Number(d)
                )
            )
            self.assertEqual(eval(expr), result)

            # piecewise(a < b, c, d)
            result = c if (a < b) else d
            expr = self.w.ex(
                Piecewise(Less(Number(a), Number(b)), Number(c), Number(d))
            )
            self.assertEqual(eval(expr), result)

            # piecewise(a <= b, c, d)
            result = c if (a <= b) else d
            expr = self.w.ex(
                Piecewise(
                    LessEqual(Number(a), Number(b)), Number(c), Number(d)
                )
            )
            self.assertEqual(eval(expr), result)

            # piecewise(a == b, c, a == d, 3, 4)
            result = c if (a == b) else (3 if (a == d) else 4)
            expr = self.w.ex(
                Piecewise(
                    Equal(Number(a), Number(b)),
                    Number(c),
                    Equal(Number(a), Number(d)),
                    Number(3),
                    Number(4),
                )
            )
            self.assertEqual(eval(expr), result)

            # piecewise(a < b, 0, c < d, 0, 5)
            result = 0 if (a < b) else (0 if (c < d) else 5)
            expr = self.w.ex(
                Piecewise(
                    Less(Number(a), Number(b)),
                    Number(0),
                    Less(Number(c), Number(d)),
                    Number(0),
                    Number(5),
                ),
            )
            self.assertEqual(eval(expr), result)


if __name__ == '__main__':
    unittest.main()



#
# Exports to SBML.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import warnings

import myokit
from myokit.formats.sbml._writer import write_file
import myokit.lib.guess
import myokit.formats.sbml as sbml


class SBMLExporter(myokit.formats.Exporter):
    """
    This :class:`Exporter<myokit.formats.Exporter>` creates an SBML model.
    """

    def __init__(self):
        super().__init__()

    def model(self, path, model, protocol=None):
        """
        Writes an SBML model to the given filename.

        Arguments:

        ``path``
            The path/filename to write the generated code to.
        ``model``
            The model to export.
        ``protocol``
            If given, an attempt will be made to convert the protocol to an
            expression and insert it into the model before exporting. See
            :meth:`myokit.lib.guess.add_embedded_protocol()` for details.

        """
        # Embed protocol
        if protocol is not None:
            model = model.clone()
            if not myokit.lib.guess.add_embedded_protocol(model, protocol):
                warnings.warn("Unable to embed stimulus protocol.")

        # Export
        sbml_model = sbml.Model.from_myokit_model(model)
        write_file(path, sbml_model)

    def supports_model(self):
        """
        Returns ``True``.
        """
        return True



#!/usr/bin/env python3
#
# Tests writing of Stan equations.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats.stan

from myokit import (
    Number, PrefixPlus, PrefixMinus, Plus, Minus,
    Multiply, Divide, Quotient, Remainder, Power, Sqrt,
    Exp, Log, Log10, Sin, Cos, Tan, ASin, ACos, ATan, Floor, Ceil, Abs,
    Not, And, Or, Equal, NotEqual, More, Less, MoreEqual, LessEqual,
    If, Piecewise,
)

import myokit.tests


class StanExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """ Test conversion of expressions to Stan. """
    _name = 'stan'
    _target = myokit.formats.stan.StanExpressionWriter

    def test_number(self):
        self.eq(Number(1), '1.0')
        self.eq(Number(-1.3274924373284374), '-1.32749243732843736e+00')
        self.eq(Number(+1.3274924373284374), '1.32749243732843736e+00')
        self.eq(Number(-2), '-2.0')
        self.eq(Number(13, 'mV'), '13.0')

    def test_name(self):
        self.eq(self.a, 'a')
        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.a), 'COMP.A')

    def test_derivative(self):
        self.eq(myokit.Derivative(self.a), 'dot(a)')

    def test_partial_derivative(self):
        e = myokit.PartialDerivative(self.a, self.b)
        self.assertRaisesRegex(NotImplementedError, 'Partial', self.w.ex, e)

    def test_initial_value(self):
        e = myokit.InitialValue(self.a)
        self.assertRaisesRegex(NotImplementedError, 'Initial', self.w.ex, e)

    def test_prefix_plus_minus(self):
        # Test with numbers
        p = Number(11, 'kV')
        self.eq(PrefixPlus(p), '+11.0')
        self.eq(PrefixPlus(PrefixPlus(p)), '++11.0')
        self.eq(PrefixMinus(p), '-11.0')
        self.eq(PrefixMinus(PrefixMinus(p)), '--11.0')
        self.eq(PrefixMinus(Number(-1)), '--1.0')

        a, b, c = self.abc
        self.eq(PrefixMinus(Plus(a, b)), '-(a + b)')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c), '+(a + b) / c')
        self.eq(Power(PrefixMinus(a), b), '(-a)^b')
        self.eq(Power(PrefixPlus(Power(b, a)), c), '(+b^a)^c')
        self.eq(Power(a, PrefixMinus(Power(b, c))), 'a^(-b^c)')

    def test_plus_minus(self):
        a, b, c = self.abc
        self.eq(Plus(a, b), 'a + b')
        self.eq(Plus(Plus(a, b), c), 'a + b + c')
        self.eq(Plus(a, Plus(b, c)), 'a + (b + c)')
        self.eq(Minus(a, b), 'a - b')
        self.eq(Minus(Minus(a, b), c), 'a - b - c')
        self.eq(Minus(a, Minus(b, c)), 'a - (b - c)')
        self.eq(Minus(a, b), 'a - b')
        self.eq(Plus(Minus(a, b), c), 'a - b + c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')
        self.eq(Minus(Plus(a, b), c), 'a + b - c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')

    def test_multiply_divide(self):
        a, b, c = self.abc
        self.eq(Multiply(a, b), 'a * b')
        self.eq(Multiply(Multiply(a, b), c), 'a * b * c')
        self.eq(Multiply(a, Multiply(b, c)), 'a * (b * c)')
        self.eq(Divide(a, b), 'a / b')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')

        self.eq(Divide(Multiply(a, b), c), 'a * b / c')
        self.eq(Multiply(Divide(a, b), c), 'a / b * c')
        self.eq(Divide(a, Multiply(b, c)), 'a / (b * c)')
        self.eq(Multiply(a, Divide(b, c)), 'a * (b / c)')

        self.eq(Multiply(Minus(a, b), c), '(a - b) * c')
        self.eq(Divide(a, Plus(b, c)), 'a / (b + c)')
        self.eq(Minus(Multiply(a, b), c), 'a * b - c')
        self.eq(Plus(a, Divide(b, c)), 'a + b / c')

    def test_quotient_remainder(self):
        a, b, c = self.abc

        self.eq(Quotient(c, a), 'floor(c / a)')
        self.eq(Remainder(c, a), '(c - a * floor(c / a))')

    def test_power(self):
        a, b, c = self.abc
        self.eq(Power(a, b), 'a^b')
        self.eq(Power(Plus(a, b), c), '(a + b)^c')
        self.eq(Power(a, Minus(b, c)), 'a^(b - c)')
        self.eq(Power(Multiply(a, b), c), '(a * b)^c')
        self.eq(Power(a, Divide(b, c)), 'a^(b / c)')

        # Stan has a right-associative power operator, so must add brackets
        # to get Myokit behavuour
        self.eq(Power(Power(a, b), c), '(a^b)^c')
        self.eq(Power(a, Power(b, c)), 'a^b^c')

    def test_functions(self):
        a, b = self.ab

        self.eq(Sqrt(a), 'sqrt(a)')
        self.eq(Exp(a), 'exp(a)')
        self.eq(Log(a), 'log(a)')
        self.eq(Log(a, b), '(log(a) / log(b))')
        self.eq(Log10(a), 'log10(a)')
        self.eq(Sin(a), 'sin(a)')
        self.eq(Cos(a), 'cos(a)')
        self.eq(Tan(a), 'tan(a)')
        self.eq(ASin(a), 'asin(a)')
        self.eq(ACos(a), 'acos(a)')
        self.eq(ATan(a), 'atan(a)')
        self.eq(Floor(a), 'floor(a)')
        self.eq(Ceil(a), 'ceil(a)')
        self.eq(Abs(a), 'abs(a)')

    def test_conditions(self):
        a, b, c, d = self.abcd

        self.eq(Equal(a, b), '(a == b)')
        self.eq(NotEqual(a, b), '(a != b)')
        self.eq(More(b, a), '(b > a)')
        self.eq(Less(d, c), '(d < c)')
        self.eq(MoreEqual(c, a), '(c >= a)')
        self.eq(LessEqual(b, d), '(b <= d)')

        self.eq(And(Equal(a, b), NotEqual(c, d)), '((a == b) && (c != d))')
        self.eq(Or(More(d, c), Less(b, a)), '((d > c) || (b < a))')
        self.eq(Not(Equal(d, d)), '(!(d == d))')

    def test_conditionals(self):
        a, b, c, d = self.abcd
        self.eq(If(Equal(a, b), d, c), '((a == b) ? d : c)')
        self.eq(Piecewise(NotEqual(d, c), b, a), '((d != c) ? b : a)')
        self.eq(Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4)),
                '((a == b) ? c : ((a == d) ? 3.0 : 4.0))')


if __name__ == '__main__':
    unittest.main()



#
# Python expression writer and keywords
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats


class PythonExpressionWriter(myokit.formats.ExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to their Python
    equivalent.
    """
    def __init__(self):
        super().__init__()

        self._flhs = None
        self.set_lhs_function(lambda v: str(v))

        self._function_prefix = 'math.'

    def set_lhs_function(self, f):
        """
        Sets a naming function, will be called to get the variable name from a
         ``myokit.LhsExpression`` object.

        The argument ``f`` should be a function that takes an ``LhsExpression``
        as input and returns a string.
        """
        self._flhs = f

    def _ex_prefix(self, e, op):
        """ Handles ex() for prefix operators. """
        # No simplifications should be made here for PrefixPlus, see
        #   https://github.com/myokit/myokit/issues/1054
        if e.bracket(e[0]):
            return f'{op}({self.ex(e[0])})'
        return f'{op}{self.ex(e[0])}'

    def _ex_infix(self, e, op):
        """ Handles ex() for infix operators, except Power. """
        if e.bracket(e[0]):
            out = f'({self.ex(e[0])}) {op}'
        else:
            out = f'{self.ex(e[0])} {op}'
        if e.bracket(e[1]):
            return f'{out} ({self.ex(e[1])})'
        else:
            return f'{out} {self.ex(e[1])}'

    def _ex_function(self, e, func):
        """ Handles ex() for function operators with operands. """
        args = ', '.join([self.ex(x) for x in e])
        return f'{self._function_prefix}{func}({args})'

    def _ex_infix_comparison(self, e, op):
        """ Handles ex() for infix comparisons (``==``, ``>=``, etc.). """
        return f'({self.ex(e[0])} {op} {self.ex(e[1])})'

    def _ex_infix_logical(self, e, op):
        """ Handles ex() for ``and`` and ``or``. """
        return f'({self.ex(e[0])} {op} {self.ex(e[1])})'

    def _ex_name(self, e):
        return self._flhs(e)

    def _ex_derivative(self, e):
        return self._flhs(e)

    def _ex_initial_value(self, e):
        raise NotImplementedError(
            'Initial values are not supported by this expression writer.')

    def _ex_partial_derivative(self, e):
        raise NotImplementedError(
            'Partial derivatives are not supported by this expression writer.')

    def _ex_number(self, e):
        return myokit.float.str(e).lstrip()

    def _ex_prefix_plus(self, e):
        return self._ex_prefix(e, '+')

    def _ex_prefix_minus(self, e):
        return self._ex_prefix(e, '-')

    def _ex_plus(self, e):
        return self._ex_infix(e, '+')

    def _ex_minus(self, e):
        return self._ex_infix(e, '-')

    def _ex_multiply(self, e):
        return self._ex_infix(e, '*')

    def _ex_divide(self, e):
        return self._ex_infix(e, '/')

    def _ex_quotient(self, e):
        return self._ex_infix(e, '//')

    def _ex_remainder(self, e):
        return self._ex_infix(e, '%')

    def _ex_power(self, e):
        # Note: Python uses a right-to-left order of operations for power, so
        # that a**b**c means a**(b**c). In Myokit, a^b^c means (a^b)^c, so we
        # need to reverse this for powers only.
        if e.bracket(e[0]) or isinstance(e[0], myokit.Power):
            out = f'({self.ex(e[0])})'
        else:
            out = f'{self.ex(e[0])}'
        if e.bracket(e[1]) and not isinstance(e[1], myokit.Power):
            return f'{out}**({self.ex(e[1])})'
        else:
            return f'{out}**{self.ex(e[1])}'

    def _ex_sqrt(self, e):
        return self._ex_function(e, 'sqrt')

    def _ex_sin(self, e):
        return self._ex_function(e, 'sin')

    def _ex_cos(self, e):
        return self._ex_function(e, 'cos')

    def _ex_tan(self, e):
        return self._ex_function(e, 'tan')

    def _ex_asin(self, e):
        return self._ex_function(e, 'asin')

    def _ex_acos(self, e):
        return self._ex_function(e, 'acos')

    def _ex_atan(self, e):
        return self._ex_function(e, 'atan')

    def _ex_exp(self, e):
        return self._ex_function(e, 'exp')

    def _ex_log(self, e):
        return self._ex_function(e, 'log')

    def _ex_log10(self, e):
        return self._ex_function(e, 'log10')

    def _ex_floor(self, e):
        return self._ex_function(e, 'floor')

    def _ex_ceil(self, e):
        return self._ex_function(e, 'ceil')

    def _ex_abs(self, e):
        return f'abs({self.ex(e[0])})'

    def _ex_equal(self, e):
        return self._ex_infix_comparison(e, '==')

    def _ex_not_equal(self, e):
        return self._ex_infix_comparison(e, '!=')

    def _ex_more(self, e):
        return self._ex_infix_comparison(e, '>')

    def _ex_less(self, e):
        return self._ex_infix_comparison(e, '<')

    def _ex_more_equal(self, e):
        return self._ex_infix_comparison(e, '>=')

    def _ex_less_equal(self, e):
        return self._ex_infix_comparison(e, '<=')

    def _ex_and(self, e):
        return self._ex_infix_logical(e, 'and')

    def _ex_or(self, e):
        return self._ex_infix_logical(e, 'or')

    def _ex_not(self, e):
        return f'(not {self.ex(e[0])})'

    def _ex_if(self, e):
        return f'({self.ex(e._t)} if {self.ex(e._i)} else {self.ex(e._e)})'

    def _ex_piecewise(self, e):
        s = ''
        n = len(e) // 2
        for i in range(0, n):
            s += f'({self.ex(e._e[i])} if {self.ex(e._i[i])} else '
        s += self.ex(e._e[n]) + ')' * n
        return s


class NumPyExpressionWriter(PythonExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to Python expressions
    intended for use in NumPy arrays.
    """
    def __init__(self):
        super().__init__()
        self._function_prefix = 'numpy.'
    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_number(self, e):
    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):
    #def _ex_quotient(self, e):
    #def _ex_remainder(self, e):
    #def _ex_power(self, e):
    #def _ex_sqrt(self, e):
    #def _ex_sin(self, e):
    #def _ex_cos(self, e):
    #def _ex_tan(self, e):

    def _ex_asin(self, e):
        return self._ex_function(e, 'arcsin')

    def _ex_acos(self, e):
        return self._ex_function(e, 'arccos')

    def _ex_atan(self, e):
        return self._ex_function(e, 'arctan')

    #def _ex_exp(self, e):

    def _ex_log(self, e):
        if len(e) == 1:
            return self._ex_function(e, 'log')
        # Always add brackets here: The parent element will have been expecting
        # a function (which never needs brackets) and so won't have added any.
        return f'(numpy.log({self.ex(e[0])}) / numpy.log({self.ex(e[1])}))'

    #def _ex_log10(self, e):
    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):

    def _ex_abs(self, e):
        # Can't use default abs here, must be numpy.abs
        return self._ex_function(e, 'abs')

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_function(e, 'logical_and')

    def _ex_or(self, e):
        return self._ex_function(e, 'logical_or')

    def _ex_not(self, e):
        return self._ex_function(e, 'logical_not')

    def _ex_if(self, e):
        return (f'{self._function_prefix}select('
                f'[{self.ex(e._i)}], [{self.ex(e._t)}], {self.ex(e._e)})')

    def _ex_piecewise(self, e):
        n = len(e._i)
        s = [self._function_prefix, 'select([']
        s.append(', '.join([self.ex(x) for x in e._i]))
        s.append('], [')
        s.append(', '.join([self.ex(x) for x in e._e[:-1]]))
        s.append('], ')
        s.append(self.ex(e._e[n]))
        s.append(')')
        return ''.join(s)




#
# C++ expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from myokit.formats.ansic import AnsiCExpressionWriter


class CppExpressionWriter(AnsiCExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    myokit :class:`expressions <myokit.Expression>` to their C++ equivalent.
    """
    # Note: The set_condition_function is used by the jacobian calculator
    # Once that has gone, this can base off the CBasedExpressionWriter without
    # any modifications (so two methods below can then go).

    def _ex_initial_value(self, e):
        raise NotImplementedError(
            'Initial values are not supported by this expression writer.')

    def _ex_partial_derivative(self, e):
        raise NotImplementedError(
            'Partial derivatives are not supported by this expression writer.')




#!/usr/bin/env python3
#
# Tests the expression writer for CUDA.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats.cuda

from myokit import (
    Number, PrefixPlus, PrefixMinus, Plus, Minus,
    Multiply, Divide, Quotient, Remainder, Power, Sqrt,
    Exp, Log, Log10, Sin, Cos, Tan, ASin, ACos, ATan, Floor, Ceil, Abs,
    Not, And, Or, Equal, NotEqual, More, Less, MoreEqual, LessEqual,
    If, Piecewise,
)

import myokit.tests


class CudaExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """
    Test conversion to Ansi C, as used by the Simulation.
    Numerical tests are provided by composing and evaluating a single RHS.
    """
    _name = 'cuda'
    _target = myokit.formats.cuda.CudaExpressionWriter

    def test_number(self):
        self.eq(Number(1), '1.0f')
        self.eq(Number(-2), '-2.0f')
        self.eq(Number(13, 'mV'), '13.0f')

    def test_number_double(self):
        w = self._target(precision=myokit.DOUBLE_PRECISION)
        self.assertEqual(w.ex(Number(1)), '1.0')
        self.assertEqual(w.ex(Number(-2)), '-2.0')
        self.assertEqual(w.ex(Number(13, 'mV')), '13.0')

    def test_name(self):
        self.eq(self.a, 'a')
        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.a), 'COMP.A')

    def test_derivative(self):
        self.eq(myokit.Derivative(self.a), 'dot(a)')

    def test_partial_derivative(self):
        self.assertRaisesRegex(
            NotImplementedError, 'Partial',
            self.w.ex, myokit.PartialDerivative(self.a, self.b))

    def test_initial_value(self):
        self.assertRaisesRegex(
            NotImplementedError, 'Initial',
            self.w.ex, myokit.InitialValue(self.a))

    def test_prefix_plus_minus(self):
        # Inherited from c-based

        p = Number(11, 'kV')
        self.eq(PrefixPlus(p), '+11.0f')
        self.eq(PrefixPlus(PrefixPlus(p)), '+(+11.0f)')
        self.eq(PrefixPlus(PrefixPlus(PrefixPlus(p))), '+(+(+11.0f))')
        self.eq(PrefixPlus(Number('+1')), '+1.0f')
        self.eq(PrefixMinus(p), '-11.0f')
        self.eq(PrefixMinus(PrefixMinus(p)), '-(-11.0f)')
        self.eq(PrefixMinus(Number(-1)), '-(-1.0f)')
        self.eq(PrefixMinus(PrefixMinus(Number(-2))), '-(-(-2.0f))')

        # Test with operators of precedence SUM, PRODUCT
        a, b, c = self.abc
        self.eq(PrefixPlus(Plus(a, b)), '+(a + b)')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c), '+(a + b) / c')
        self.eq(PrefixPlus(Divide(b, a)), '+(b / a)')
        self.eq(PrefixMinus(Minus(a, b)), '-(a - b)')
        self.eq(Multiply(PrefixMinus(Plus(b, a)), c), '-(b + a) * c')
        self.eq(PrefixMinus(Divide(b, a)), '-(b / a)')

    def test_prefix_plus_minus_double(self):
        # Inherited from c-based

        w = self._target(precision=myokit.DOUBLE_PRECISION)
        p = Number(3, 'mA')
        self.assertEqual(w.ex(PrefixPlus(p)), '+3.0')
        self.assertEqual(w.ex(PrefixPlus(PrefixPlus(p))), '+(+3.0)')
        self.assertEqual(
            w.ex(PrefixPlus(PrefixPlus(PrefixPlus(p)))), '+(+(+3.0))')
        self.assertEqual(w.ex(PrefixPlus(Number('+1'))), '+1.0')
        self.assertEqual(w.ex(PrefixMinus(p)), '-3.0')
        self.assertEqual(w.ex(PrefixMinus(PrefixMinus(p))), '-(-3.0)')
        self.assertEqual(w.ex(PrefixMinus(Number(-1))), '-(-1.0)')
        self.assertEqual(
            w.ex(PrefixMinus(PrefixMinus(Number(-2)))), '-(-(-2.0))')

    def test_arithmetic(self):
        # Inherited from c-based

        a, b, c = self.abc
        self.eq(Minus(Plus(a, b), c), 'a + b - c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')
        self.eq(Multiply(Minus(a, b), c), '(a - b) * c')
        self.eq(Multiply(a, Plus(b, c)), 'a * (b + c)')
        self.eq(Minus(Multiply(a, b), c), 'a * b - c')
        self.eq(Divide(a, Minus(b, c)), 'a / (b - c)')
        self.eq(Plus(Divide(a, b), c), 'a / b + c')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')

    def test_quotient_remainder(self):
        # Inherited from c-based

        a, b, c = self.abc
        self.eq(Quotient(a, Divide(b, c)), 'floorf(a / (b / c))')
        self.eq(Remainder(Plus(a, c), b), '(a + c - b * floorf((a + c) / b))')
        self.eq(Divide(c, Remainder(b, a)), 'c / ((b - a * floorf(b / a)))')

    def test_quotient_remainder_double(self):
        w = self._target(precision=myokit.DOUBLE_PRECISION)
        a, b, c = self.abc
        self.assertEqual(w.ex(Quotient(a, b)), 'floor(comp.a / comp.b)')

    def test_power(self):
        a, b, c = self.abc
        self.eq(Power(a, b), 'powf(a, b)')
        self.eq(Power(Power(a, b), c), 'powf(powf(a, b), c)')
        self.eq(Power(a, Power(b, c)), 'powf(a, powf(b, c))')

        self.eq(Power(Plus(a, b), c), 'powf(a + b, c)')
        self.eq(Power(a, Minus(b, c)), 'powf(a, b - c)')
        self.eq(Power(Multiply(a, b), c), 'powf(a * b, c)')
        self.eq(Power(a, Divide(b, c)), 'powf(a, b / c)')

    def test_power_double(self):
        w = self._target(precision=myokit.DOUBLE_PRECISION)
        a, b, c = self.abc
        self.assertEqual(w.ex(Power(a, b)), 'pow(comp.a, comp.b)')
        self.assertEqual(w.ex(Power(Plus(a, b), c)),
                         'pow(comp.a + comp.b, comp.c)')

    def test_log(self):
        a, b = self.ab
        self.eq(Log(a), 'logf(a)')
        self.eq(Log10(a), 'log10f(a)')
        self.eq(Log(a, b), '(logf(a) / logf(b))')

    def test_log_double(self):
        w = self._target(precision=myokit.DOUBLE_PRECISION)
        a, b = self.ab
        self.assertEqual(w.ex(Log(a)), 'log(comp.a)')
        self.assertEqual(w.ex(Log10(b)), 'log10(comp.b)')
        self.assertEqual(w.ex(Log(b, a)), '(log(comp.b) / log(comp.a))')

    def test_functions(self):
        a = self.a
        self.eq(Sqrt(self.a), 'sqrtf(a)')
        self.eq(Exp(self.a), 'expf(a)')
        self.eq(Sin(self.a), 'sinf(a)')
        self.eq(Cos(self.a), 'cosf(a)')
        self.eq(Tan(self.a), 'tanf(a)')
        self.eq(ASin(self.a), 'asinf(a)')
        self.eq(ACos(self.a), 'acosf(a)')
        self.eq(ATan(self.a), 'atanf(a)')
        self.eq(Floor(self.a), 'floorf(a)')
        self.eq(Ceil(self.a), 'ceilf(a)')
        self.eq(Abs(self.a), 'fabsf(a)')

    def test_functions_double(self):
        w = self._target(precision=myokit.DOUBLE_PRECISION)
        self.assertEqual(w.ex(Sqrt(self.a)), 'sqrt(comp.a)')
        self.assertEqual(w.ex(Exp(self.a)), 'exp(comp.a)')
        self.assertEqual(w.ex(Sin(self.a)), 'sin(comp.a)')
        self.assertEqual(w.ex(Cos(self.a)), 'cos(comp.a)')
        self.assertEqual(w.ex(Tan(self.a)), 'tan(comp.a)')
        self.assertEqual(w.ex(ASin(self.a)), 'asin(comp.a)')
        self.assertEqual(w.ex(ACos(self.a)), 'acos(comp.a)')
        self.assertEqual(w.ex(ATan(self.a)), 'atan(comp.a)')
        self.assertEqual(w.ex(Floor(self.a)), 'floor(comp.a)')
        self.assertEqual(w.ex(Ceil(self.a)), 'ceil(comp.a)')
        self.assertEqual(w.ex(Abs(self.a)), 'fabs(comp.a)')

    def test_conditions(self):
        # Inherited from c-based

        a, b, c, d = self.abcd
        p = Equal(a, b)
        q = NotEqual(c, d)
        self.eq(And(p, q), '((a == b) && (c != d))')
        self.eq(Or(q, p), '((c != d) || (a == b))')
        self.eq(Not(And(p, p)), '(!((a == b) && (a == b)))')
        self.eq(Not(q), '(!(c != d))')

        self.eq(Equal(a, b), '(a == b)')
        self.eq(NotEqual(a, b), '(a != b)')
        self.eq(More(b, a), '(b > a)')
        self.eq(Less(d, c), '(d < c)')
        self.eq(MoreEqual(c, a), '(c >= a)')
        self.eq(LessEqual(b, d), '(b <= d)')

        self.eq(Or(More(d, c), Less(b, a)), '((d > c) || (b < a))')
        self.eq(Not(Equal(Number(1), Number(2))), '(!(1.0f == 2.0f))')
        self.eq(Not(Less(Number(1), Number(2))), '(!(1.0f < 2.0f))')

    def test_conditionals(self):
        # Inherited from c-based

        a, b, c, d = self.abcd
        self.eq(If(Equal(a, b), d, c), '((a == b) ? d : c)')
        self.eq(Piecewise(NotEqual(d, c), b, a), '((d != c) ? b : a)')
        self.eq(Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4)),
                '((a == b) ? c : ((a == d) ? 3.0f : 4.0f))')


if __name__ == '__main__':
    unittest.main()



#
# OpenCL expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.ansic import CBasedExpressionWriter


class OpenCLExpressionWriter(CBasedExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to OpenCL syntax.

    Arguments:

    ``precision``
        By default, numbers are shown as e.g. ``1.23f``, denoting single
        precision literals. To use double precision instead, set ``precision``
        to ``myokit.DOUBLE_PRECISION``.
    ``native_math``
        By default, the software implementations of functions like ``log`` and
        ``exp`` are used. To use the native version instead, set
        ``native_math`` to ``True``.

    """
    def __init__(self, precision=myokit.SINGLE_PRECISION, native_math=True):
        super().__init__()
        self._sp = (precision == myokit.SINGLE_PRECISION)
        self._nm = bool(native_math)

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_initial_value(self, e):
    #def _ex_partial_derivative(self, e):

    def _ex_number(self, e):
        x = super()._ex_number(e)
        return x + 'f' if self._sp else x

    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):

    # Native divide seemed to cause some issues
    #def _ex_divide(self, e):
    #    if self._nm:
    #        return 'native_divide(' + self.ex(e[0]) +', '+ self.ex(e[1]) + ')'
    #    return self._ex_infix(e, '/')

    #def _ex_quotient(self, e):
    #def _ex_remainder(self, e):
    #def _ex_power(self, e):

    def _ex_sqrt(self, e):
        f = 'native_sqrt' if self._nm else 'sqrt'
        return self._ex_function(e, f)

    def _ex_sin(self, e):
        f = 'native_sin' if self._nm else 'sin'
        return self._ex_function(e, f)

    def _ex_cos(self, e):
        f = 'native_cos' if self._nm else 'cos'
        return self._ex_function(e, f)

    def _ex_tan(self, e):
        f = 'native_tan' if self._nm else 'tan'
        return self._ex_function(e, f)

    #def _ex_asin(self, e):
    #def _ex_acos(self, e):
    #def _ex_atan(self, e):

    def _ex_exp(self, e):
        f = 'native_exp' if self._nm else 'exp'
        return self._ex_function(e, f)

    def _ex_log(self, e):
        log = 'native_log' if self._nm else 'log'
        if len(e) == 1:
            return self._ex_function(e, log)
        # Always add brackets: parent was expecting a function so will never
        # have added them.
        return f'({log}({self.ex(e[0])}) / {log}({self.ex(e[1])}))'

    def _ex_log10(self, e):
        f = 'native_log10' if self._nm else 'log10'
        return self._ex_function(e, f)

    #def _ex_floor(self, e):
    #def _ex_ceil(self, e):
    #def _ex_abs(self, e):

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    #def _ex_and(self, e):
    #def _ex_or(self, e):
    #def _ex_not(self, e):
    #def _ex_if(self, e):
    #def _ex_piecewise(self, e):




#!/usr/bin/env python3
#
# Tests writing of Stan equations.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats.latex

from myokit import (
    Number, PrefixPlus, PrefixMinus, Plus, Minus,
    Multiply, Divide, Quotient, Remainder, Power, Sqrt,
    Exp, Log, Log10, Sin, Cos, Tan, ASin, ACos, ATan, Floor, Ceil, Abs,
    Not, And, Or, Equal, NotEqual, More, Less, MoreEqual, LessEqual,
)

import myokit.tests


class LatexExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """
    Test conversion of expressions to Python.
    This is used by pyfunc(), although that's usually done with the
    NumPyExpressionWriter instead.
    Numerical tests are provided.
    """
    _name = 'latex'
    _target = myokit.formats.latex.LatexExpressionWriter
    _update_lhs_function = False

    def test_number(self):
        self.eq(Number(1), '1.0')
        self.eq(Number(-1.3274924373284374), '-1.32749243732843736e+00')
        self.eq(Number(+1.3274924373284374), '1.32749243732843736e+00')
        self.eq(Number(-2), '-2.0')
        self.eq(Number(13, 'mV'), r'13.0 \text{mV}')
        self.eq(Number(2, 'A/F'), r'2.0 \text{A/F}')
        self.eq(Number(7, myokit.units.dimensionless), '7.0')

    def test_name(self):
        self.eq(self.a, r'\text{a}')
        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('under_score', rhs=3)
        t = c.add_variable('time', rhs=0, binding='time')
        m.validate()
        self.eq(myokit.Name(v), r'\text{under\_score}')

        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.a), 'COMP.A')

    def test_derivative(self):
        self.eq(myokit.Derivative(self.a), r'\frac{d\text{a}}{d\text{t}}')

        w = self._target()
        w.set_time_variable_name('timmy')
        self.assertEqual(w.ex(myokit.Derivative(self.a)),
                         r'\frac{d\text{a}}{d\text{timmy}}')
        w.set_time_variable_name('Bob')
        self.assertEqual(w.ex(myokit.Derivative(self.a)),
                         r'\frac{d\text{a}}{d\text{Bob}}')

    def test_partial_derivative(self):
        self.eq(myokit.PartialDerivative(self.a, self.b),
                r'\frac{\partial\text{a}}{\partial\text{b}}')

    def test_initial_value(self):
        self.eq(myokit.InitialValue(self.a), r'\text{a}(\text{t} = 0)')

    def test_prefix_plus(self):
        # Test with numbers
        p = Number(11, 'mV')
        self.eq(PrefixPlus(p), r'+11.0 \text{mV}')
        p = Number(3)
        self.eq(PrefixPlus(PrefixPlus(p)), '++3.0')
        self.eq(PrefixPlus(Number('+1')), '+1.0')

        # Test with operators of precedence SUM, PRODUCT, POWER
        a, b, c = self.abc
        self.eq(PrefixPlus(Plus(a, b)), r'+\left(\text{a}+\text{b}\right)')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c),
                r'\frac{+\left(\text{a}+\text{b}\right)}{\text{c}}')
        self.eq(Power(PrefixPlus(b), a),
                r'\left(+\text{b}\right)^\text{a}')

    def test_prefix_minus(self):
        # Test with numbers
        p = Number(3, 'uA')
        self.eq(PrefixMinus(p), r'-3.0 \text{uA}')
        p = Number(2)
        self.eq(PrefixMinus(PrefixMinus(p)), '--2.0')
        self.eq(PrefixMinus(Number('-1')), '--1.0')

        # Test with operators of precedence SUM, PRODUCT, POWER
        a, b, c = self.abc
        self.eq(PrefixMinus(Plus(a, b)), r'-\left(\text{a}+\text{b}\right)')
        self.eq(Divide(PrefixMinus(Plus(a, b)), c),
                r'\frac{-\left(\text{a}+\text{b}\right)}{\text{c}}')
        self.eq(Power(PrefixMinus(b), a),
                r'\left(-\text{b}\right)^\text{a}')

    def test_plus_minus(self):
        a, b, c = self.abc
        ta, tb, tc = r'\text{a}', r'\text{b}', r'\text{c}'
        self.eq(Plus(a, b), f'{ta}+{tb}')
        self.eq(Plus(Plus(a, b), c), f'{ta}+{tb}+{tc}')
        self.eq(Plus(a, Plus(b, c)), rf'{ta}+\left({tb}+{tc}\right)')

        self.eq(Minus(a, b), f'{ta}-{tb}')
        self.eq(Minus(Minus(a, b), c), f'{ta}-{tb}-{tc}')
        self.eq(Minus(a, Minus(b, c)), rf'{ta}-\left({tb}-{tc}\right)')

        self.eq(Minus(a, b), f'{ta}-{tb}')
        self.eq(Plus(Minus(a, b), c), f'{ta}-{tb}+{tc}')
        self.eq(Minus(a, Plus(b, c)), rf'{ta}-\left({tb}+{tc}\right)')
        self.eq(Minus(Plus(a, b), c), f'{ta}+{tb}-{tc}')
        self.eq(Minus(a, Plus(b, c)), rf'{ta}-\left({tb}+{tc}\right)')

    def test_multiply_divide(self):
        a, b, c = self.abc
        ta, tb, tc = r'\text{a}', r'\text{b}', r'\text{c}'
        l, r = r'\left(', r'\right)'
        self.eq(Multiply(a, b), rf'{ta}\cdot{tb}')
        self.eq(Multiply(Multiply(a, b), c), rf'{ta}\cdot{tb}\cdot{tc}')
        self.eq(Multiply(a, Multiply(b, c)), rf'{ta}\cdot{l}{tb}\cdot{tc}{r}')
        self.eq(Divide(a, b), r'\frac{\text{a}}{\text{b}}')
        self.eq(Divide(Divide(a, b), c),
                r'\frac{\frac{\text{a}}{\text{b}}}{\text{c}}')
        self.eq(Divide(a, Divide(b, c)),
                r'\frac{\text{a}}{\frac{\text{b}}{\text{c}}}')
        self.eq(Divide(Divide(a, b), c),
                r'\frac{\frac{\text{a}}{\text{b}}}{\text{c}}')
        self.eq(Divide(Multiply(a, b), c),
                r'\frac{\text{a}\cdot\text{b}}{\text{c}}')

        self.eq(Multiply(Minus(a, b), c), rf'{l}{ta}-{tb}{r}\cdot{tc}')
        self.eq(Multiply(a, Plus(b, c)), rf'{ta}\cdot{l}{tb}+{tc}{r}')
        self.eq(Plus(a, Multiply(b, c)), rf'{ta}+{tb}\cdot{tc}')

    def test_remainder_quotient(self):
        a, b = self.ab
        self.eq(Remainder(a, b), r'\text{a}\bmod\text{b}')
        self.eq(Quotient(a, b),
                r'\left\lfloor\frac{\text{a}}{\text{b}}\right\rfloor')

    def test_power(self):
        a, b, c = self.abc
        ta, tb, tc = r'\text{a}', r'\text{b}', r'\text{c}'
        l, r = r'\left(', r'\right)'
        p, q = '{', '}'
        self.eq(Power(a, b), f'{ta}^{tb}')
        self.eq(Power(Power(a, b), c), f'{l}{ta}^{tb}{r}^{tc}')
        self.eq(Power(a, Power(b, c)), f'{ta}^{p}{tb}^{tc}{q}')
        self.eq(Power(Plus(a, b), c), f'{l}{ta}+{tb}{r}^{tc}')
        self.eq(Power(a, Minus(b, c)), f'{ta}^{p}{tb}-{tc}{q}')

    def test_functions(self):
        a, b = self.ab

        self.eq(Sqrt(a), r'\sqrt{\text{a}}')
        self.eq(Exp(a), r'\exp\left(\text{a}\right)')
        self.eq(Log(a), r'\log\left(\text{a}\right)')
        self.eq(Log(a, b), r'\log_{\text{b}}\left(\text{a}\right)')
        self.eq(Log10(a), r'\log_{10}\left(\text{a}\right)')
        self.eq(Sin(a), r'\sin\left(\text{a}\right)')
        self.eq(Cos(a), r'\cos\left(\text{a}\right)')
        self.eq(Tan(a), r'\tan\left(\text{a}\right)')
        self.eq(ASin(a), r'\arcsin\left(\text{a}\right)')
        self.eq(ACos(a), r'\arccos\left(\text{a}\right)')
        self.eq(ATan(a), r'\arctan\left(\text{a}\right)')
        self.eq(Floor(a), r'\left\lfloor{\text{a}}\right\rfloor')
        self.eq(Ceil(a), r'\left\lceil{\text{a}}\right\rceil')
        self.eq(Abs(a), r'\lvert{\text{a}}\rvert')

    def test_conditions(self):
        a, b, c, d = self.abcd

        self.eq(Equal(a, b), r'\left(\text{a}=\text{b}\right)')
        self.eq(NotEqual(a, b), r'\left(\text{a}\neq\text{b}\right)')
        self.eq(More(b, a), r'\left(\text{b}>\text{a}\right)')
        self.eq(Less(d, c), r'\left(\text{d}<\text{c}\right)')
        self.eq(MoreEqual(c, a), r'\left(\text{c}\geq\text{a}\right)')
        self.eq(LessEqual(b, d), r'\left(\text{b}\leq\text{d}\right)')

        self.eq(And(Equal(a, b), NotEqual(c, d)),
                r'\left(\left(\text{a}=\text{b}\right)\and'
                r'\left(\text{c}\neq\text{d}\right)\right)')
        self.eq(Or(Equal(a, b), NotEqual(c, d)),
                r'\left(\left(\text{a}=\text{b}\right)\or'
                r'\left(\text{c}\neq\text{d}\right)\right)')
        self.eq(Not(Equal(d, d)),
                r'\left(\not\left(\text{d}=\text{d}\right)\right)')

    def test_conditionals(self):
        a, b, c, d = self.abcd
        p = myokit.Equal(a, b)
        q = myokit.NotEqual(c, d)
        self.eq(myokit.If(p, c, d), r'\text{if}\left(\left(\text{a}=\text{b}'
                r'\right),\text{c},\text{d}\right)')
        self.eq(myokit.Piecewise(p, b, q, d, Number(1)), r'\text{piecewise}'
                r'\left(\left(\text{a}=\text{b}\right),\text{b},\left(\text{c}'
                r'\neq\text{d}\right),\text{d},1.0\right)')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests shared infrastructure in myokit.formats.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.formats

from myokit.tests import TemporaryDirectory


class FormatsTest(unittest.TestCase):
    """ Test shared formats functionality. """

    def test_register_external_importer(self):
        # Tests registering an external importer.

        class TestImporter(myokit.formats.Importer):
            pass

        self.assertRaisesRegex(
            KeyError, 'Importer not found',
            myokit.formats.importer, 'testo')

        myokit.formats.register_external_importer('testo', TestImporter)
        e = myokit.formats.importer('testo')
        self.assertIsInstance(e, TestImporter)
        e = myokit.formats.importer('testo')
        self.assertIsInstance(e, TestImporter)

        # Test overwriting
        myokit.formats.register_external_importer('testo', TestImporter)
        e = myokit.formats.importer('testo')
        self.assertIsInstance(e, TestImporter)

        # Test removing
        myokit.formats.register_external_importer('testo', None)
        self.assertRaisesRegex(
            KeyError, 'Importer not found',
            myokit.formats.importer, 'testo')

    def test_register_external_exporter(self):
        # Tests registering an external exporter.

        class TestExporter(myokit.formats.Exporter):
            pass

        self.assertRaisesRegex(
            KeyError, 'Exporter not found',
            myokit.formats.exporter, 'testo')

        myokit.formats.register_external_exporter('testo', TestExporter)
        e = myokit.formats.exporter('testo')
        self.assertIsInstance(e, TestExporter)
        e = myokit.formats.exporter('testo')
        self.assertIsInstance(e, TestExporter)

        # Test overwriting
        myokit.formats.register_external_exporter('testo', TestExporter)
        e = myokit.formats.exporter('testo')
        self.assertIsInstance(e, TestExporter)

        # Test removing
        myokit.formats.register_external_exporter('testo', None)
        self.assertRaisesRegex(
            KeyError, 'Exporter not found',
            myokit.formats.exporter, 'testo')

    def test_register_external_ewriter(self):
        # Tests registering an external expression writer.

        class TestExpressionWriter(myokit.formats.ExpressionWriter):
            pass

        self.assertRaisesRegex(
            KeyError, 'Expression writer not found',
            myokit.formats.ewriter, 'testo')

        myokit.formats.register_external_ewriter('testo', TestExpressionWriter)
        e = myokit.formats.ewriter('testo')
        self.assertIsInstance(e, TestExpressionWriter)
        e = myokit.formats.ewriter('testo')
        self.assertIsInstance(e, TestExpressionWriter)

        # Test overwriting
        myokit.formats.register_external_ewriter('testo', TestExpressionWriter)
        e = myokit.formats.ewriter('testo')
        self.assertIsInstance(e, TestExpressionWriter)

        # Test removing
        myokit.formats.register_external_ewriter('testo', None)
        self.assertRaisesRegex(
            KeyError, 'Expression writer not found',
            myokit.formats.ewriter, 'testo')


class ExporterTest(unittest.TestCase):
    """ Tests shared :class:`Exporter` functionality. """

    def test_writable_dir(self):
        # Test :meth:`_test_writable_dir` for existing paths.

        m, p, x = myokit.load('example')
        e = myokit.formats.exporter('ansic')
        with TemporaryDirectory() as d:

            # Create in new dir
            path = d.path('new-dir')
            e.runnable(path, m, p)

            # Create again, in same dir (should be ok)
            e.runnable(path, m, p)

            # Create at location of file: not allowed!
            path = d.path('file')
            with open(path, 'w') as f:
                f.write('Hello')
            self.assertRaisesRegex(
                myokit.ExportError, 'file exists', e.runnable, path, m, p)

    def test_runnable_exporter(self):
        # Test shared functionality of the TemplatedRunnableExporters.

        e = myokit.formats.exporter('ansic')

        # Load model, protocol
        m, p, x = myokit.load('example')

        # Create empty output directory as subdirectory of DIR_OUT
        with TemporaryDirectory() as d:
            path = d.path()

            # Simple export
            dpath = os.path.join(path, 'runnable1')
            ret = e.runnable(dpath, m)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # Write to complex path
            dpath = os.path.join(path, 'runnable2', 'nest', 'test')
            ret = e.runnable(dpath, m, p)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # Overwrite existing path
            ret = e.runnable(dpath, m, p)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # Path pointing to file
            dpath = os.path.join(path, 'file')
            with open(dpath, 'w') as f:
                f.write('contents\n')
            self.assertRaisesRegex(
                myokit.ExportError, 'file exists', e.runnable, dpath, m, p)

            # Directory exists where we're trying to write a file
            dpath = os.path.join(path, 'runnable3')
            fname = os.path.join(dpath, 'sim.c')
            os.makedirs(fname)
            self.assertRaisesRegex(
                myokit.ExportError, 'Directory exists',
                e.runnable, dpath, m, p)

            # Directory embedded in the output file path
            def embedded():
                return {'sim.c': 'nested/sim.c'}

            # 1. Normal operation
            e._dict = embedded
            dpath = os.path.join(path, 'runnable4')
            ret = e.runnable(dpath, m, p)
            self.assertIsNone(ret)
            self.assertTrue(os.path.isdir(dpath))
            self.assertTrue(len(os.listdir(dpath)) > 0)

            # 2. Try to create directory where file exists
            def embedded():
                return {'sim.c': 'nested/sim.c/som.c'}

            e._dict = embedded
            dpath = os.path.join(path, 'runnable4')
            self.assertRaisesRegex(
                myokit.ExportError, 'file or link', e.runnable, dpath, m, p)


class ImporterTest(unittest.TestCase):
    """ Test shared importer functionality. """

    def test_importer_interface(self):
        # Test listing and creating importers.
        ims = myokit.formats.importers()
        self.assertTrue(len(ims) > 0)
        for i in ims:
            self.assertIsInstance(i, str)
            i = myokit.formats.importer(i)
            self.assertTrue(isinstance(i, myokit.formats.Importer))

    def test_unknown(self):
        # Test requesting an unknown importer.
        # Test fetching using importer method
        self.assertRaisesRegex(
            KeyError, 'Importer not found', myokit.formats.importer, 'blip')


class EWriterTest(unittest.TestCase):
    """ Test shared ewriter functionality. """

    def test_ewriter_interface(self):
        # Test listing and creating expression writers.

        # Test listing
        es = myokit.formats.ewriters()
        self.assertTrue(len(es) > 0)

        # Create one of each
        for e in es:
            self.assertIsInstance(e, str)
            e = myokit.formats.ewriter(e)
            self.assertTrue(isinstance(e, myokit.formats.ExpressionWriter))

    def test_unknown(self):
        # Test requesting an unknown expression writer.
        # Test fetching using ewriter method
        self.assertRaisesRegex(
            KeyError, 'Expression writer not found', myokit.formats.ewriter,
            'dada')


if __name__ == '__main__':
    unittest.main()



#
# Provides DiffSL support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._ewriter import DiffSLExpressionWriter
from ._exporter import DiffSLExporter

# Importers

# Exporters
_exporters = {
    'diffsl': DiffSLExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'diffsl': DiffSLExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


#
# Language keywords
#
keywords = [
    'abs',
    'cos',
    'dudt',
    'exp',
    'F',
    'G',
    'heaviside',
    'in',
    'log',
    'M',
    'out',
    'pow',
    'sigmoid',
    'sin',
    'sqrt',
    't',
    'tan',
    'u',
]
